from loguru import logger
from typing import List

TOKEN_CUT = {
    ("alphanum", "alphanum"): False,  # word
    ("alphanum", "space"): True,  # end of word
    ("alphanum", "dot"): False,  # dotsep
    ("alphanum", "other"): True,  # end of word
    ("space", "alphanum"): True,  # word
    ("space", "space"): None,  # end of word
    ("space", "dot"): True,  # number
    ("space", "other"): True,  # end of word
    ("dot", "alphanum"): False,  # word
    ("dot", "space"): True,  # end of word
    ("dot", "dot"): False,  # number
    ("dot", "other"): True,  # end of word
    ("other", "alphanum"): True,  # word
    ("other", "space"): True,  # end of word
    ("other", "dot"): True,  # number
    ("other", "other"): True,  # end of word
}

META_CHARS = {
    ".eq.": "\xaa",  # comparison
    ".neq.": "\xab",
    ".ge.": "\xac",
    ".gt.": "\xad",
    ".le.": "\xae",
    ".lt.": "\xaf",
    ".and.": "\xba",  # logic
    ".or.": "\xbb",
    "//": "\xca",  # markers
    "/*": "\xcb",
    "*/": "\xcc",
    "::": "\xcd",
    "||": "\xce",
    "&&": "\xcf",
    "+=": "\xda",  # operands
    "-=": "\xdb",
    "/=": "\xdc",
    "*=": "\xdd",
    "**": "\xde",
    "==": "\xdf",
    ">=": "\xea",  # operands
    "<=": "\xeb",
    "<<": "\xec",
}

META_CHARS_REV = {value: key for key, value in META_CHARS.items()}

STRING_START = "\xfa"
STRING_STOP = "\xfb"
STRING_SUB = "\xfc"

STR_SUBSTITUTE = {
    "\xfa": '"',
    "\xfb": '"',
    "\xfc": "'",
}


# def string_encode(line: str) -> str:
#     string_type = ""
#     new_line = ""
#     for char in line:
#         if char in ["'", '"']:
#             if string_type == "":  # <- Enter string
#                 new_line += STRING_START
#                 string_type = char
#             else:  # in a string
#                 if char == string_type:
#                     new_line += STRING_STOP
#                     string_type = ""
#                 else:
#                     new_line += STRING_SUB
#         else:
#             new_line += char
#     assert len(new_line) == len(line)
#     return new_line

def string_encode(line: str) -> str:
    string_type = ""
    result = []
    for char in line:
        if char in ["'", '"']:
            if not string_type:  # Enter string
                result.append(STRING_START)
                string_type = char
            elif char == string_type:  # Close string
                result.append(STRING_STOP)
                string_type = ""
            else:  # Different string character within string
                result.append(STRING_SUB)
        else:
            result.append(char)
    return ''.join(result)

# def string_decode(line: str) -> str:
#     new_line = ""
#     for char in line:
#         new_line += STR_SUBSTITUTE.get(char, char)
#     return new_line

def string_decode(line: str) -> str:
    return ''.join(STR_SUBSTITUTE.get(char, char) for char in line)


def metachar_encode(line: str) -> str:
    for char, proxy in META_CHARS.items():
        line = line.replace(char, proxy)
    return string_encode(line)


# def metachar_encode(line: str) -> str:
#     # cannot work, its not by single char
#     return string_encode(''.join(META_CHARS.get(char, char) for char in line))


# def metachar_decode(line: str) -> str:
#     for char, proxy in META_CHARS.items():
#         line = line.replace(proxy, char)
#     return string_decode(line)

# def metachar_decode(line: str) -> str:
#     result = line
#     for char, proxy in META_CHARS.items():
#         result = result.replace(proxy, char)
#     return string_decode(result)


def metachar_decode(line: str) -> str:
    return string_decode(''.join(META_CHARS_REV.get(char, char) for char in line))



# def metachar_decode_from_list(list_: str) -> List:
#     # logger.critical(list_)
#     # list_ = [  META_CHARS_REV.get(item, item) for item in list_]
#     list_ = [metachar_decode(item) for item in list_]

#     # logger.critical(list_)
#     list_ = [string_decode(item) for item in list_]
#     # logger.critical(list_)
#     return list_

def metachar_decode_from_list(list_: List[str]) -> List[str]:
    return [string_decode(metachar_decode(item)) for item in list_]

def iterate_with_neighbors(iterable_):
    # Create iterators for the current and previous characters
    prev = None  # Initialize previous character to None
    for curr in iterable_:
        # Yield the previous and current characters
        yield prev, curr
        prev = curr
    yield prev, None


def tokenize(line: str) -> List[str]:
    """
    Light tokenizer to ease code identification
    """

    def _cast(char):
        if char == " ":
            return "space"
        if char in [".", "_", "%", "\xcd"]:  # \xcd is the :: marker
            return "dot"
        if char in META_CHARS_REV.keys():
            return "other"
        if char.isalnum():
            return "alphanum"
        return "other"

    tokens = []
    buffer = ""
    prev_cast = "other"
    instring = False
    for curr_char in metachar_encode(line):
        cur_cast = _cast(curr_char)
        to_cut = TOKEN_CUT[prev_cast, cur_cast]

        # string override; a string is a single token
        if curr_char == STRING_START:
                instring = True
                if buffer.strip() != "":
                    tokens.append(buffer)
                buffer = ""
        if curr_char == STRING_STOP:
            if buffer.strip() != "":
                buffer += curr_char
                tokens.append(buffer)
            buffer = ""
            instring = False
            continue

        if instring:
            buffer += curr_char
            continue
        # string override; a string is a single token

        if to_cut is None:
            continue
        elif to_cut is True:
            if buffer.strip() != "":
                tokens.append(buffer)
            buffer = curr_char
        else:
            buffer += curr_char

        prev_cast = cur_cast

    if buffer.strip() != "":
        tokens.append(buffer)

    return metachar_decode_from_list(tokens)


def find_words_before_left_parenthesis_noregexp(tokens: List[str]) -> List[str]:
    """Find all words before a left parenthesis in a line"""
    if "(" not in tokens:
        return []
    matches = []
    for prev, curr in iterate_with_neighbors(tokens):
        if curr == "(":
            try:
                if prev not in ",+-/*<>=;|(){}[]:&!~ ":
                    matches.append(prev)
            except TypeError:
                pass  # triggered by prev == None

    clean_matches = sorted(set(matches))

    black_list = list(META_CHARS.keys()) + ["!"]
    # remove meta chars , whit ar not words...
    clean_matches = [item for item in clean_matches if item not in black_list]

    return clean_matches


def get_indent(line: str) -> str:
    """Get the indentation leading a line"""
    _indent = ""
    for char in line:
        if char == "\t":
            _indent += "    "
        elif char != " ":
            return _indent
        else:
            _indent += " "
    return _indent


def average_indent(code:List[str])->float:
    """Average indentation lvl of code"""
    sum=0
    for line in code:
        sum+=len(get_indent(line))
    return 1.*sum / len(code)


def eat_spaces(code: List[str]) -> List[str]:
    """Remove unwanted multiple spacing"""
    new_stmt = []
    for line in code:
        line = line.replace("\t","  ") 
        out = get_indent(line)

        prevchar = None
        for i, char in enumerate(line.strip()):
            try:
                next_char = line.strip()[i + 1]
            except IndexError:
                next_char = None

            if char == " ":
                if prevchar not in [" ", ":", ";", ","] and next_char not in [
                    ":",
                    ";",
                    ",",
                ]:
                    out += char
                else:
                    pass  # no space needed if " " precedes, or a punctuation is before or after
            else:
                out += char

            prevchar = char
        new_stmt.append(out)
    return new_stmt


def get_common_root_index(list_:list)-> int:
    "return the index of the first character differing after a common root"
    if "" in list_:
        return 0


    def root_ok_str(list_:list, root):
        """ 
        - list_ is a list of str and root is a string
        - list_ is a list of list of str and root is a list of str
        
        """
        for item_ in list_:
            if not item_.startswith(root) :
                return False
        return True
    

    one_ = list_[0]
    idx = 0
    while root_ok_str(list_, one_[:idx]):# and idx<len(list_[0])-1 :
        #logger.info(one_[:idx])
        #logger.info(idx)
        idx+=1
        
    out = idx-2
            
    return out



def get_common_root_index_list(list_:List[List[str]])-> int:
    "return the index of the first character differing after a common root"
    if [] in list_:
        return 0


    def root_ok_list(list_:list, root):
        """ 
        - list_ is a list of str and root is a string
        - list_ is a list of list of str and root is a list of str
        
        """
        #logger.warning(f"root {root}")
        #logger.warning(f"all items {list_}")
        
        for item_ in list_:
            #logger.warning(f"root {root}, item {item_[0:len(root)]}" )
            if root != item_[0:len(root)] :
                return False
        return True
    

    one_ = list_[0]
    idx = 0
    while root_ok_list(list_, one_[:idx]):
        #logger.info(one_[:idx])
        #logger.info(idx)
        idx+=1
        
    out = idx-2
    #logger.info(out)
            
    return out



def best_front_match(str_ :iter,list_str:List[iter])->iter:
    """
    Return the str in LIST_STR with the longest match w.r. to str_
    generalized to iterables if needed.
    """
    
    #references = [ref for ref in references if ref.startswith(prefix)]
    lmatch =0
    smatch=None
    for candidate in list_str:
        count = front_match(str_,candidate)
        if count > lmatch:
            lmatch =count
            smatch = candidate    
    return smatch

def front_match(list1: iter, list2: iter) -> int:
    """
    Count the number of matching elements from the end of two iterable.

    Args:
        list1 (iterable): The first iterable to compare.
        list2 (iterable): The second iterable to compare.

    Returns:
        int: The number of matching elements from the end of both iterable.
    """
    count = 0

    # Compare elements from the end of the lists
    for elem1, elem2 in zip(list1, list2):
        if elem1 == elem2:
            count += 1
        else:
            break

    return count


def inner_match(list1: iter, list2: iter) -> int:
    """
    Count the number of sequential matching elements between two iterable.

    Args:
        list1 (iterable): The first iterable to compare.
        list2 (iterable): The second iterable to compare.

    Returns:
        int: The number of matching elements from the end of both iterable.
    """
    count = 0

    # Compare elements from the end of the lists
    jstart = None
    for istart, item in enumerate(list1):
        if item in list2:
            jstart=list2.index(item)
            break
    
    if jstart is None:
        #print(f"None of {list1} found in {list2}")
        return 0

    count = 0
    try:
        while list1[istart+count] == list2[jstart+count]:
            count+=1
    except IndexError:
        pass 
    #print(f"Found {count} matchs for {list1} and {list2}")
    return count

def best_inner_match(str_ :iter,list_str:List[iter])->list:
    """
    Return the str in LIST_STR with the longest match w.r. to str_
    generalized to iterables if needed.
    """
    
    #references = [ref for ref in references if ref.startswith(prefix)]
    lmatch =0
    imatch=[]
    for i,candidate in enumerate(list_str):
        count = inner_match(str_,candidate)
        #logger.info(f"Found {count} matchs for {str_} in {candidate}")
        if count == 0:
            continue
        if count > lmatch:
            lmatch =count
            imatch=[i]
        elif count == lmatch:
            imatch.append(i)  
    return imatch


def generate_dict_paths(d:dict, path:list=None)-> list:
    if path is None:
        path = []

    if isinstance(d, dict) and d != {} :
        paths = []
        for k, v in d.items():
            new_path = path + [k]
            paths.extend(generate_dict_paths(v, new_path))
        return paths
    else:
        return [path]





from pathlib import Path
from loguru import logger


def rec_travel_through_package(
    path: str,
    optional_paths: list = None,
) -> list:
    """
    List all paths from a folder and its sub-folders recursively.

    RECURSIVE
    """
    if not optional_paths:
        optional_paths = []

    current_paths_list = [path, *optional_paths]

    # Move to absolute Path objects
    current_paths_list = [Path(p_) for p_ in current_paths_list]

    accepted_suffixes = ["f","f90","f77", "c","cc", "cpp","h","hpp","py"]

    paths_ = []
    for current_path in current_paths_list:
        for element in current_path.iterdir():
            if element.is_dir():
                path_str = element.as_posix()
                paths_.extend(rec_travel_through_package(path_str))
            else:
                if element.as_posix().split(".")[-1].lower() not in accepted_suffixes:
                    continue
                if element.as_posix() not in paths_:
                    if not element.as_posix().split("/")[-1].startswith("."):
                        paths_.append(element.as_posix())
    return paths_


def clean_extensions_in_paths(paths_list: list) -> list:
    """
    Remove unwanted path extensions and duplicates.

    Args:
        paths_list (list): List of all paths gatheres through recursive analysis

    Returns:
        list: List of cleaned paths.
    """
    clean_paths = []
    for path in paths_list:
        if path.endswith(
            (
                ".py",
                ".f",
                ".F",
                ".f77",
                ".f90",
                ".F77",
                ".F90",
                ".c",
                ".cpp",
                ".h",
                ".hpp",
            )
        ):
            clean_paths.append(path)

    return [
        *set(clean_paths),
    ]


def get_package_files(clean_paths: list, relpath: str) -> dict:
    """
    Return all the files useful for a package analysis, with their absolut paths
    """

    files = []
    for path_ in clean_paths:
        if not Path(path_).is_dir():
            #            logger.info(f"Append :{path_}")
            files.append(path_)

    files = clean_extensions_in_paths(files)

    if not files:
        logger.warning(f"No files found in the paths provided")

    files = [Path(relpath) / Path(p_) for p_ in files]

    out = {}
    for file in files:
        path_ = file.relative_to(Path(relpath)).as_posix()
        out[path_] = file.as_posix()
    return out



"""Custom tucan exeptions"""


class TucanError(RuntimeError):
    """General tucan error
    
    If you are coming from another package, use this by default """
    pass

class TucanParsingError(TucanError):
    """raised is tucan reach a known potential dead en in parsing code
    
    e.g. reaching an fortran end statement withoud corresponding context"""
    pass

class TucanCtrlPtsError(TucanError):
    """raised is tucan reach a dead end when building the control points
    
    e.g. interpreting python code with spurious indentations"""
    pass


"""
Module that gather the most common functions of struct.
"""

from dataclasses import dataclass
from typing import Tuple, List, Callable
from copy import deepcopy
from loguru import logger
from typing import List
from tucan.complexities import maintainability_index
from tucan.string_utils import tokenize, get_indent

def path_clean(path: list, paths_to_clean: Tuple[list]) -> list:
    """Remove the unwanted steps of the paths"""
    indexes_to_clean = []
    for ptc in paths_to_clean:
        if list2pathref(path).startswith(list2pathref(ptc)):
            indexes_to_clean.append(len(ptc) - 1)
    new_path = []
    for i, step in enumerate(path):
        if i not in indexes_to_clean:
            new_path.append(step)
    return new_path


def list2pathref(path: list) -> str:
    """The way we will refer to path here in strings"""
    return ".".join(path)


def pathref_ascendants(pathstr: str) -> List[str]:
    """Return all ascendends of a path"""
    out = []
    path = pathstr.split(".")
    while len(path) > 1:
        path.pop(-1)
        out.append(list2pathref(path))
    return out



def show_time(timesec:int)->str:
    out = str(timesec) + " seconds"
    if timesec >= 60:
        out = f"{int(timesec/60):4g} min"
    if timesec >= 3600:
        out = f"{int(timesec/3600):4g} hrs"
    if timesec >= 3600*24:
        out = f"{(int(timesec/(3600*24))):4g}days"
    return out



def struct_summary_str(main_structs: dict) -> str:
    out = []

    if main_structs is None:
        return "No structure data is available. Tucan light parser could not interpret this file."
    for part, data in main_structs.items():
        out.append(f'\n{data["type"]} {part} :')
        out.append(
            f'    At path {data["path"]}, name {data["name"]}, lines {data["lines"][0]} -> {data["lines"][-1]}'
        )
        out.append(f'    {data["ssize"]} statements over {data["NLOC"]} lines')
        out.append("    >>> Complexity <<<")
        
        out.append(
            f'    Mcabe ctrlPts  CCN: {data["CCN"]:8d} (inside),   {data["CCN_int"]:8g} (CCN_int intensive)'
        )
        out.append(
            f'    PossiblePaths  CPP: {data["CPP"]:8g} (inside),   {data["CPP_int"]:8g} (CPP_int intensive)'
        )
        out.append(
            f'    Halstead Diff. HDF: {data["HDF"]:8g} (inside),   {data["HDF_int"]:8g} (HDF_int intensive)'
        )
        out.append(
            f'    Avg. Indents.  IDT: {data["IDT"]:8g} (inside),   {data["IDT_int"]:8g} (IDT_int intensive)'
        )
        out.append(
            f'    Struct Cplx.   CST: {data["CST"]:8g} (inside),   {data["CST_ext"]:8g} (CST_ext extensive)'
        )
        out.append(
            f'    Maintn. Index  MI : {data["MI"]:8g} (inside),   {data["MI_int"]:8g} ( MI_int intensive)'
        )
        out.append(
            f'    Halstead Time. HTM: {show_time(data["HTM"])} (inside),   {show_time(data["HTM_ext"])} (HTM_ext extensive)'
        )
        
        
        if data["callables"]:
            list_str = "\n       - " + "\n       - ".join(data["callables"])

            out.append(f'    Refers to {len(data["callables"])} callables:{list_str}')

        if data["contains"]:
            list_str = "\n    - " + "\n    - ".join(data["contains"])
            out.append(f'    Contains {len(data["contains"])} elements:{list_str}')

        if data["parents"]:
            list_str = "\n       - " + "\n       - ".join(data["parents"])
            out.append(f'    Refers to {len(data["parents"])} parents:{list_str}')

        if data["annotations"]:
            keyvals = [
                key + ":" + values for key, values in data["annotations"].items()
            ]
            list_str = "\n       - " + "\n       - ".join(keyvals)

            out.append(f"    Refers to {len(keyvals)} callables:{list_str}")

    return "\n".join(out)


########################################################
# BUFFER of detection
@dataclass
class BufferItem:
    """Forces buffers to keep the same logic across languages"""

    type_: str = None
    name: str = None
    path: list = None
    first_line: str = None
    line_idx: int = None
    statement_idx: int = None
    parents: List[str] = None
    callables: List[str] = None
    contains: List[str] = None
    comment: str = None


def new_buffer_item(
    type_: str = None,
    name: str = None,
    path: List[str] = None,
    first_line: str = None,
    line_idx: int = None,
    statement_idx: int = None,
    verbose: bool = False,
    parents: List[str] = None,
    callables: List[str] = None,
    contains: List[str] = None,
    comment: str = None,
) -> BufferItem:
    if verbose:
        fname = ".".join(path)
        logger.critical(f"START l.{line_idx} for " + fname)
    if parents is None:
        parents = []
    if callables is None:
        callables = []
    if contains is None:
        contains = []
    if comment is None:
        comment = ""
    out = BufferItem(
        type_=type_,
        name=name,
        path=path,
        first_line=first_line,
        line_idx=line_idx,
        statement_idx=statement_idx,
        parents=parents,
        callables=callables,
        contains=contains,
        comment=comment,
    )
    return out


########################################################
# STACK of detection
@dataclass
class StackItem:
    """Forces buffers to keep the same logic across languages"""

    type_: str
    name: str
    path: list
    start_line_idx: int
    start_statement_idx: int
    start_line: str
    end_line_idx: int
    end_statement_idx: int
    end_line: str
    parents: List[str] = None
    callables: List[str] = None
    contains: List[str] = None
    comment: str = None


def new_stack_item(
    buf: BufferItem,
    end_line_idx: int,
    end_statement_idx: int,
    end_line: str,
    verbose: bool = False,
) -> StackItem:
    if verbose:
        fname = ".".join(buf.path)
        logger.critical(f" END   l.{end_line_idx} for " + fname + '|'+end_line.strip())
    out = StackItem(
        type_=buf.type_,
        name=buf.name,
        path=buf.path.copy(),
        start_line_idx=buf.line_idx,
        start_statement_idx=buf.statement_idx,
        start_line=buf.first_line,
        parents=buf.parents,
        callables=buf.callables,
        contains=buf.contains,
        comment=buf.comment,
        end_line_idx=end_line_idx,
        end_statement_idx=end_statement_idx,
        end_line=end_line,
    )
    return out


def struct_from_stack(stack: list, main_types: list, skip_types: list = None) -> dict:
    """Build a dictionary of all structures"""
    # Build nested structure
    struct = {}
    if skip_types is None:
        skip_types = []

    path_to_skip = []
    for stack_item in stack:
        if stack_item.type_ in skip_types:
            path_to_skip.append(stack_item.path)

    for stack_item in stack:
        cleaned_path = path_clean(stack_item.path, path_to_skip)
        if stack_item.type_ in main_types:
            # logger.warning(f"Adding {list2pathref(cleaned_path)}")
            id_ = list2pathref(cleaned_path)
            if id_ not in struct:
                struct[id_] = {
                    "path": cleaned_path,
                    "name": stack_item.name,
                    "type": stack_item.type_,
                    "linestart": stack_item.start_line,
                    "lines": [stack_item.start_line_idx, stack_item.end_line_idx],
                    "statements": [
                        stack_item.start_statement_idx,
                        stack_item.end_statement_idx,
                    ],  # Warning: here statements starts at 1!!!
                    "contains": stack_item.contains,
                    "parents": stack_item.parents,
                    "callables": stack_item.callables,
                    "comment": stack_item.comment,
                    "annotations": {},
                }
            else: # create a proxy because this structure is redefined
                id_new = id_ + f"#{stack_item.start_line_idx},{stack_item.end_line_idx}"
                struct[id_new] = {
                    "path": cleaned_path,
                    "name": stack_item.name,
                    "type": stack_item.type_,
                    "linestart": stack_item.start_line,
                    "lines": [stack_item.start_line_idx, stack_item.end_line_idx],
                    "statements": [
                        stack_item.start_statement_idx,
                        stack_item.end_statement_idx,
                    ],  # Warning: here statements starts at 1!!!
                    "contains": stack_item.contains,
                    "parents": stack_item.parents,
                    "callables": stack_item.callables,
                    "comment": stack_item.comment,
                    "annotations": {},
                }
                struct[id_]["contains"].append(id_new)
            
           

    return struct


def get_struct_sizes(struct: dict) -> dict:
    """Compute the size of strict items (statefull)"""
    struct_aspects = {}
    for part, data in struct.items():
        struct_aspects[part] = {}
        struct_aspects[part]["NLOC"] = data["lines"][-1] - data["lines"][0] + 1
        struct_aspects[part]["ssize"] = data["statements"][-1] - data["statements"][0]
        
    return struct_aspects


def replace_self(list_: list, parent: str) -> list:
    """Replace the self keyword in a parentality path"""
    return [item.replace("self.", parent + ".") for item in list_]


def _strip_safe_lines(beg: int, end: int, safes: List[list]) -> List:
    """Return an iterable stripped from safe zones
    beg=100
    end = 110
    safes = [[103,104],[106,109]]

    100
    101
    102
    105

    """
    iter_ = []
    for i in range(beg, end + 1):
        blocked = False
        for safe in safes:
            if i >= safe[0] and i <= safe[1]:
                # print(f"{i} blocked")
                blocked = True
        if not blocked:
            iter_.append(i)
    return iter_


def struct_actual_lines(struct_in: dict, name: str) -> list:
    """returns an iterable with only the statement relative to this part
    excluding contained parts.

    WARNING:The -1 on statements is systematic because statements numbering is starting at 1
    """
    data = struct_in[name]
    safes = []
    for sub_name in data["contains"]:
        try:
            safes.append(
                [
                    struct_in[sub_name]["statements"][0] - 1,
                    struct_in[sub_name]["statements"][1] - 1,
                ]
            )
        except KeyError:
            msgerr = f"Item {sub_name} is not referenced in this context"
            raise RuntimeError(msgerr)

    return _strip_safe_lines(
        data["statements"][0] - 1, data["statements"][1] - 1, safes
    )


def struct_augment(
    struct_in: dict,
    clean_code: List[str],
    find_callables: Callable,
    compute_complexities: Callable,
    compute_cst: Callable,
) -> dict:
    """Complete the description of each struct item"""
    struct = deepcopy(struct_in)
    # first lines computation
    for _, data in struct.items():
        data["NLOC"] = data["lines"][-1] - data["lines"][0] + 1
        data["ssize"] = data["statements"][-1] - data["statements"][0] + 1

    # add internal links
    for part, data in struct.items():
        path = data["path"]
        # logger.warning(path)

        if len(path) > 1:
            parent = path[:-1] + path[-1].split(".")[:-1]
            try:
                struct[list2pathref(parent)]["contains"].append(list2pathref(path))
                # pass
            except KeyError:
                pass
                # will happen for scripts, with "dummy" not always referenced.
            # struct[part]["parents"].append(list2pathref(parent))
        # else:
        #     struct[part]["parent"]=None

    # add language specific analyses
    for part, data in struct.items():
        actual_lines = struct_actual_lines(struct, part)
        sub_code = [clean_code[i] for i in actual_lines]
       # logger.critical(part)
        # for i,line in enumerate(clean_code):
        #     if i  in actual_lines:
        #         logger.success(line)
        #     else:
        #         logger.warning(line)

        sub_tokenized_code = [tokenize(line) for line in sub_code[1:]]
        sub_indents_code=  [ len(get_indent(line)) for line in sub_code[1:]]

       
        data["callables"].extend(find_callables(sub_tokenized_code))
        if data["parents"]:
            data["callables"] = replace_self(data["callables"], data["parents"][0])
        if data["type"] in ["class"]:
            data["contains"] = replace_self(data["contains"], part)
            data["callables"] = replace_self(data["callables"], part)

        if len(sub_tokenized_code)==0:
            avg_indent=0
            mccabe=1
            possible_paths=1
            volume=0
            difficulty=0
            time_to_code=0
        else:
            avg_indent,mccabe,possible_paths,volume, difficulty,time_to_code=compute_complexities(sub_indents_code,sub_tokenized_code)
        data["CCN"] = mccabe
        data["IDT"] = avg_indent
        data["CPP"] = possible_paths
        data["HDF"] = difficulty
        data["HTM"] = time_to_code
        data["CST"] = compute_cst(data["type"])
        data["MI"] = maintainability_index(volume,mccabe, data["ssize"])
        
    



    fields_intensive = ["CCN","CPP","HDF","MI","IDT"]
    fields_extensive = ["HTM","CST"]
    def recursive_aggregate(struct, label):
        if "aggregated" not in struct[label]:
            sums_ext = {field:0 for field in fields_extensive}
            sums_int = {field:0 for field in fields_intensive}

            for child in struct[label]["contains"]:
                recursive_aggregate(struct, child)
                sums_ext = {field: sums_ext[field]+struct[child][field]  for field in fields_extensive}
                sums_int = {field: sums_int[field]+struct[child][field]*struct[child]["ssize"]  for field in fields_intensive}

            for field in fields_extensive:
                struct[label][field+"_ext"]= sums_ext[field]+ struct[label][field]
            for field in fields_intensive:
                struct[label][field+"_int"]= round(sums_int[field] /struct[label]["ssize"] + struct[label][field],2)

            struct[label]["aggregated"]=True

                
    for part in struct.keys():
        if "aggregated" not in struct[part]:
            recursive_aggregate(struct, part)


    return struct



"""Global function to handle the unformat of various languages"""
from loguru import logger

from tucan.guess_language import guess_language
from tucan.unformat_common import read_lines_with_encoding
from tucan.unformat_py import unformat_py
from tucan.unformat_ftn import unformat_ftn
from tucan.unformat_c import unformat_c
from tucan.clean_ifdef import remove_ifdef_from_module


def unformat_main(filename: str, verbose: bool = False) -> list:
    """
    Main function to call to get an unformated version of the code

    Args:
        filename (str): _description_

    Returns:
        list: _description_
    """
    logger.info(f"Unformatting {filename}")

    code = read_lines_with_encoding(filename)

    code = [line.lower() for line in code]  # Lower case for all

    if filename.lower().endswith(".py"):
        logger.debug(f"Python code detected ...")
        statements = unformat_py(code)
    elif filename.lower().endswith((".f", ".F", ".f77", ".F77", ".f90", ".F90")):
        logger.debug(f"Fortran code detected ...")
        code = remove_ifdef_from_module(code, [], verbose, fortran=True)
        statements = unformat_ftn(code)
    elif filename.lower().endswith((".c", ".cpp", ".cc")):
        logger.debug(f"C/C++ code detected ...")
        code = remove_ifdef_from_module(code, [], verbose)
        statements = unformat_c(code)
    elif filename.lower().endswith((".h", ".hpp")):
        lang = guess_language(code)
        if lang in ["ccpp"]:
            logger.debug(f"C/C++ code detected ...")
            code = remove_ifdef_from_module(code, [], verbose=False)
            statements = unformat_c(code)
        elif lang in ["fortran"]:
            logger.debug(f"Fortran code detected ...")
            code = remove_ifdef_from_module(code, [], verbose=False, fortran=True)
            statements = unformat_ftn(code)
        else:
            logger.debug(f"Language was not either C or Fortran, skipping...")
            statements = None
    else:
        ext = filename.lower().split(".")[-1]
        logger.error(f"Extension {ext} not supported, exiting ...")
        statements = None

    return statements



from typing import List
from loguru import logger

from tucan.unformat_common import (
    Statements,
    new_stmts,
    remove_strings,
    clean_blanks,
    clean_inline_comments,
    rm_parenthesis_content,
    align_multiline_blocks,
    align_end_continuations,
)

from tucan.string_utils import eat_spaces, get_indent, tokenize

from tucan.kw_lang import KEYWORDS_C


# TODO remove all spaces and reformat?
def remove_space_in_front_of_variables(stmts: Statements) -> Statements:
    """_summary_

    Args:
        stmts (Statements): _description_

    Returns:
        Statements: _description_
    """
    new_stmt = []
    new_lines = []
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        stmt = line
        for keyword in KEYWORDS_C:
            if keyword in line.split() and "=" in line:
                if line.split()[1] == "=":
                    stmt = line.replace(line.split()[0] + " =", line.split()[0] + "=")
                    logger.warning(
                        f"A C Keywords {keyword} is used as a variable in the code. Bad Practice Should Be Avoided"
                    )
                continue
            continue

        new_stmt.append(stmt)
        new_lines.append([lstart, lend])

    return Statements(new_stmt, new_lines)


def split_blocks(stmts: Statements) -> Statements:
    """Split statements on { and } markers"""
    new_stmt = []
    new_lines = []
    stmt = ""
    level = 0
    INDENT = "  "
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        # skip void lines
        if line.strip() == "":
            continue

        # exception for includes
        if line.startswith("#"):
            new_stmt.append(line)
            new_lines.append([lstart, lend])
            continue

        for char in line:
            if char in ["{"]:
                if stmt.strip() == "":  # edge case, line directly starts with {
                    new_stmt[-1] += (
                        " " + char
                    )  # we add this char to the previous statement
                    new_lines[-1][
                        -1
                    ] = lend  # and update the endline of prev statement  (ONLY CASE Where linespan changes)
                else:
                    new_stmt.append(
                        level * INDENT + stmt.lstrip() + char
                    )  # normal case , we add the statement with the char
                    new_lines.append([lstart, lend])  # and the lines spa
                level += 1  # lvl is increased
                stmt = ""  # stmt is reset
            elif char in ["}"]:
                if stmt.strip() != "":  # edge case, line directly starts with }
                    new_stmt.append(
                        level * INDENT + stmt.lstrip()
                    )  # we add the last statement  wo the "}"
                    new_lines.append([lstart, lend])  # same line span
                level -= 1  # lvl is increased
                new_stmt.append(
                    level * INDENT + char
                )  # a new statement is added with "}"
                new_lines.append([lstart, lend])  # same line span
                stmt = ""  # stmt is reset
            elif (
                char in [";"]
                and new_stmt
                and new_stmt[-1][-1] in ["}"]
                and stmt.strip() == ""
            ):
                continue  # if a ; follows directly a }
            else:
                stmt += char

        # line is finished. If statmt is not void, add sttment
        if stmt.strip() != "":
            new_stmt.append(level * INDENT + stmt.lstrip())
            new_lines.append([lstart, lend])
            stmt = ""

    return Statements(new_stmt, new_lines)


def merge_lines_by_end(stmts: Statements) -> Statements:
    """Ensure lines only ends with {,},;,:"""
    new_stmt = []
    new_lines = []
    stmt = ""
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if stmt == "":
            tmp_start = lstart
            stmt = get_indent(line)
        else:
            stmt += " "

        stmt += line.lstrip()

        if line.startswith("#"):
            new_stmt.append(stmt)
            new_lines.append([tmp_start, lend])
            stmt = ""
            continue

        if line.endswith(("{", ";", ":", "}")) or lend == stmts.lines[-1][-1]:
            new_stmt.append(stmt)
            new_lines.append([tmp_start, lend])
            stmt = ""

    return Statements(new_stmt, new_lines)


def merge_parenthesis_lines(stmts: Statements) -> Statements:
    new_stmt = []
    new_lines = []
    stmt = ""
    lvl = 0
    istart = 0
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if lvl > 0:
            line = " " + line.lstrip()
        else:
            istart = lstart

        for char in line:
            if char in ["("]:
                lvl += 1
            if char in [")"]:
                lvl -= 1
            stmt += char

        if lvl == 0:
            new_stmt.append(stmt)
            new_lines.append([istart, lend])
            stmt = ""

    return Statements(new_stmt, new_lines)


# def split_multiple_statements(stmts: Statements) -> Statements:
#     """Split lines with multiples ; """
#     new_stmt = []
#     new_lines = []
#     for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
#         if line.lstrip().startswith("for "):
#             new_stmt.append(line)
#             new_lines.append([lstart, lend])
#             continue

#         stmt = line.rstrip(";").rstrip()
#         indent=get_indent(line)
#         for item in stmt.split(";"):
#             new_stmt.append(indent+item.strip()+";")
#             new_lines.append([lstart, lend])

#     return Statements(new_stmt, new_lines)


# def split_declarations(stmts: Statements) -> Statements:
# """Split functions declarations, to make clean signatures """
# new_stmt = []
# new_lines = []
# tail = " ###==============================="
# def _need_split(buffer:str)->bool:
# try:
# cleanstr = rm_parenthesis_content(buffer[:-1])
# type_,name_ = cleanstr.lstrip().split()
# if "=" not in name_:
# return True
# except ValueError:
# pass
# return False


# for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
# buffer=""
# for char in line:
# buffer+=char
# if char == "{":
# if _need_split(buffer):
# new_stmt.append(buffer+tail)
# new_lines.append([lstart, lend])
# buffer=get_indent(line)+"    "
# if char == "}":
# pass
# new_stmt.append(buffer)
# new_lines.append([lstart, lend])
# return Statements(new_stmt, new_lines)


def line_to_flag(line, iline):
    """Check if line should be flagged"""
    # obvious fallbacks

    tokens = tokenize(line)

    if tokens[0] in ["template"]:
        return True

    if not tokens[-1] in ["{"]:
        return False

    if tokens[0] in ["if", "case", "else", "switch", "enum", "return"]:
        return False

    if "=" in tokens:
        return False

    if len(tokens) <= 2:
        return False

    #    if not tokens[1].isalnum():
    #        return False

    #   logger.critical(tokens)

    return True


def flag_declarations(code: List[str]) -> List[str]:
    """Split functions declarations, to make clean signatures"""
    new_lines = []
    tail = " ###==============================="

    for iline, line in enumerate(code):
        new_lines.append(line)
        if line_to_flag(line, iline):
            if line.endswith(";"):
                new_lines[-1] += "{}"
            new_lines[-1] += tail

    return new_lines


def unformat_c(code: List[str]) -> Statements:
    """
    Unformat C code by stripping comments and moving leading '&' characters.

    Args:
        code (List[str]): List of C code lines.

    Returns:
        List[Tuple[str, Tuple[int, int]]]: List of unformatted code statements along with line number ranges.
    """
    stmts = new_stmts(code)
    stmts.stmt = eat_spaces(stmts.stmt)
    stmts.stmt = remove_strings(stmts.stmt, '"')
    stmts.stmt = remove_strings(stmts.stmt, "'")
    stmts = align_multiline_blocks(stmts, "/*", "*/")
    stmts = clean_inline_comments(
        stmts, "/*"
    )  # this ine must follow the align multiline block
    stmts = clean_inline_comments(stmts, "//")
    stmts = clean_blanks(
        stmts
    )  # this one must follow the clean inline, to remove blank lines
    stmts = align_end_continuations(stmts, "\\")
    stmts = split_blocks(stmts)
    stmts = merge_parenthesis_lines(stmts)
    stmts = merge_lines_by_end(stmts)
    stmts.stmt = flag_declarations(stmts.stmt)

    # stmts = remove_space_in_front_of_variables(stmts)

    return stmts



"""Global function to handle the struct analysis of various languages"""
from loguru import logger

from tucan.unformat_common import read_lines_with_encoding
from tucan.unformat_py import unformat_py
from tucan.unformat_ftn import unformat_ftn
from tucan.unformat_c import unformat_c
from tucan.struct_py import extract_struct_py
from tucan.struct_ftn import extract_struct_ftn
from tucan.struct_c import extract_struct_c
from tucan.clean_ifdef import remove_ifdef_from_module
from tucan.guess_language import guess_language


def struct_main(filename: str, verbose: bool = False) -> dict:
    """
    Extract structure of a fortran or python file.
    - Find the nested structures of a code
    - Find the callables in each structure
    - Evaluate sizes, CCN

    Args:
        filename (str): Name of the file (with its path) to parse.

    Returns:
        dict: Structure analyzed, with complexity, size, name, path, lines span, etc.

        void dictionary if failed.

    """
    logger.info(f"Struct analysis on {filename}")
    code = read_lines_with_encoding(filename)

    code = [line.lower() for line in code]  # Lower case for all

    if filename.lower().endswith(".py"):
        logger.debug(f"Python code detected ...")
        statements = unformat_py(code)
        struct_ = extract_struct_py(statements, verbose)
    elif filename.lower().endswith((".f", ".F", ".f77", ".F77", ".f90", ".F90")):
        logger.debug(f"Fortran code detected ...")
        code = remove_ifdef_from_module(code, [], verbose=False, fortran=True)
        statements = unformat_ftn(code)
        struct_ = extract_struct_ftn(statements, verbose)
    elif filename.lower().endswith((".c", ".cpp", ".cc")):
        logger.debug(f"C/C++ code detected ...")
        code = remove_ifdef_from_module(code, [], verbose=False)
        statements = unformat_c(code)
        struct_ = extract_struct_c(statements, verbose)
    elif filename.lower().endswith((".h", ".hpp")):
        lang = guess_language(code)
        logger.info(f"Language of .h file is ({lang}), ")

        if lang in ["fortran"]:
            logger.debug(f"Fortran code detected ...")
            code = remove_ifdef_from_module(code, [], verbose=False, fortran=True)
            statements = unformat_ftn(code)
            struct_ = extract_struct_ftn(statements, verbose)
        else:  # lang in ["ccpp"]:
            logger.debug(f"C/C++ code detected ...")
            code = remove_ifdef_from_module(code, [], verbose=False)
            statements = unformat_c(code)
            struct_ = extract_struct_c(statements, verbose)

        # else:
        #     raise RuntimeError()

    else:
        ext = filename.lower().split(".")[-1]
        logger.error(f"Extension {ext} not supported, exiting ...")
        return {}

    return struct_


def create_empty_struct(filename: str) -> dict:
    """
    Function to create an empty structure output.

    Args:
        filename (str): Name of the file (with its path) to parse.

    Returns:
        dict: Empty structure dict i.e. with defaults values.
    """
    struct_ = {
        filename: {
            "CCN": 1,
            "NLOC": 1,
            "callables": [],
            "contains": [],
            "lines": [0, 0],
            "linestart": None,
            "name": filename,
            "path": [filename],
            "ssize": 0,
            "statements": [],
            "type": None,
        }
    }
    return struct_



import re
from typing import Tuple, List
from loguru import logger
from tucan.unformat_common import Statements
from tucan.struct_common import (
    new_buffer_item,
    new_stack_item,
    struct_from_stack,
    struct_augment,
)
from tucan.unformat_common import rm_parenthesis_content
from tucan.string_utils import find_words_before_left_parenthesis_noregexp, tokenize
from tucan.tucanexceptions import TucanParsingError,TucanCtrlPtsError

from tucan.kw_lang import STRUCTURES_FTN, NESTS_FTN, OTHERS_FTN
from tucan.complexities import (
    halstead_numbers,
    halstead_properties,
    compute_cyclomatic_complexity,
    compute_possible_paths,
)


PARTS = sorted(STRUCTURES_FTN, reverse=True)
NESTS = sorted(NESTS_FTN, reverse=True)
INTRINSICS = OTHERS_FTN


def extract_struct_ftn(stmts: Statements, verbose) -> dict:
    """Main calls to build structure form statements

    statements is the output of tucan.unformat_ftn.unformat_ftn
    """
    clean_code = stmts.to_code()

    all_structs = _extract_on_cleaned_ftn(stmts, verbose=verbose)
    all_structs = struct_augment(
        all_structs,
        clean_code,
        find_callables_ftn,
        compute_complexities_ftn,
        compute_cst_ftn,
    )
    return all_structs


def _extract_on_cleaned_ftn(stmts: Statements, verbose=False) -> dict:
    """Extract structure from cleaned statements."""
    buffer_pile = []
    stack = []
    path = []

    ins_ = []
    out_ = []
    stat_idx = 0
    for line, (line_idx1, line_idx2) in zip(stmts.stmt, stmts.lines):
        stat_idx += 1

        line = line.lstrip(" ")

        # you must also note nests because bare end statement can jam the system
        # kw are listed in reverse order to try longer kwords first: type_real before
        # for part in sorted(STRUCTURES_FTN + NESTS_FTN, reverse=True):
        for part in sorted(PARTS + NESTS, reverse=True):
            if line.startswith(part + " ") or line == part:
                ins_.append(line)
                name, type_ = _parse_name_ftn(line, line_idx1)
                path.append(name)
                buffer_pile.append(
                    new_buffer_item(
                        type_=type_,
                        path=path,
                        name=name,
                        first_line=line,
                        line_idx=line_idx1,
                        statement_idx=stat_idx,
                        verbose=verbose,
                    )
                )
                break  # must not look at the following part, or both function and function_elemental will trigger

        if "procedure" in line:
            tokens = tokenize(line)
            if "=>" in tokens:
                idx = tokens.index("=>")
                target = tokens[idx + 1]
                proxy = tokens[idx - 1]
            else:
                target = tokens[-1]
                proxy = tokens[-1]

            target_name = ".".join(path[:-1]) + "." + target
            proxy_name = ".".join(path) + "." + proxy
            path.append(proxy)
            last_buff = buffer_pile[-1]

            # logger.warning(f"Adding proxy {proxy_name} (pointer to {target_name}) to {last_buff.name}")

            buf_ = new_buffer_item(
                type_="procedure",
                path=path,
                name=proxy_name,
                first_line=line,
                line_idx=line_idx1,
                statement_idx=stat_idx,
                verbose=verbose,
                callables=[target_name],
            )
            stack.append(
                new_stack_item(buf_, line_idx1, stat_idx, line, verbose=verbose)
            )
            path.pop(-1)

        if line.startswith("__type_from__"):
            parent = tokenize(line)[2]
            # logger.warning(f"Adding parent {parent}")
            if parent not in buffer_pile[-1].parents:
                buffer_pile[-1].parents.append(parent)

        if line.startswith("type") and "extends" in line:
            strt_ = line.index("extends")
            read = False
            parent = ""
            for char in line[strt_:]:
                if char == "(":
                    read = True
                elif char == ")":
                    break
                else:
                    if read:
                        parent += char
            parent = parent.strip()
            # logger.warning(f"Adding parent {parent}")
            if parent not in buffer_pile[-1].parents:
                buffer_pile[-1].parents.append(parent)

        if line.startswith("end ") or line.strip() == "end":
            
            if not buffer_pile:
                msg = f"No code read matching END statement {line_idx1}:{line}"
                logger.critical(msg)
                raise TucanParsingError(msg)
            
            last_buff = buffer_pile[-1]  # Here
            
            items=line.split()
            if len(items)>=2:
                expected_type = last_buff.type_
                if expected_type == "select_type": # warning : not  tested because not sure we should rely to this select_type keyword
                    expected_type="select"
                elif expected_type == "select_case": # warning : not  tested because not sure we should rely to this select_case keyword
                    expected_type="select"
                type_end=items[1]
                if expected_type != type_end:
                    msg = f"Bad end type for END statement  {line_idx1}:{line} (expecting {expected_type})"
                    logger.critical(msg)
                    raise TucanParsingError(msg)

            stack.append(
                new_stack_item(last_buff, line_idx2, stat_idx, line, verbose=verbose)
            )
            path.pop(-1)
            buffer_pile.pop(-1)
            continue

    # Check specific to fortran
    for stack_item in stack:
        short_type = stack_item.type_.split("_")[0].strip()
        if short_type not in stack_item.end_line:
            pathstr = ".".join(path)
            logger.debug(
                f"End mismatch \nat {pathstr} for {short_type}:\n '{stack_item.start_line_idx}' to '{stack_item.end_line_idx}'.\n For {stack_item.type_} in {stack_item.end_line}"
            )

    # if len(ins_) != len(out_):
    #     for aa, bb in zip(ins_, out_):
    #         logger.warning(f"{aa}<>{bb}")
    #     logger.error(
    #         "Missing one structure statement such as end if... removing file from current analysis"
    #     )
    #     return {}

    return struct_from_stack(stack, main_types=PARTS + ["procedure"])


def _parse_name_ftn(line: str, line_nb: int) -> Tuple[str, str]:
    """expect a lowercase stripped line
    takes the second word as the name
    """

    line = rm_parenthesis_content(line)

    tokens = tokenize(line)
    # no names
    if tokens[0] in NESTS + [
        "interface"
    ]:  # because interface is usually used without name
        name = line.split()[0] + str(line_nb)
        if "#LABEL" in line:
            name += "_" + line.split("#")[-1].split(":")[-1].strip()
    elif "::" in tokens:
        idx = tokens.index("::")
        name = tokens[idx + 1]
    else:
        name = tokens[1]

    type_ = tokens[0]

    return name, type_


##### FTN specific functions


def find_callables_ftn(tokenized_code: List[list]) -> list:
    """Find callables in fortran"""
    candidates = []
    for tokens in tokenized_code:
        if "call" in tokens:
            candidates.append(tokens[tokens.index("call")+1])
        else:
            candidates.extend(find_words_before_left_parenthesis_noregexp(tokens))
    # NB we expect lines like 'call mysubroutine()' to be caught by left parenthesis law$
    matches = [
        cand.replace("%", ".")
        for cand in set(candidates)
        if cand
        not in INTRINSICS
        + [
            "__type_is__",
            "__type_from__",
            "select_type",
            "double_precision",
            "if",
            ".or.",
            ".and.",
        ]
    ]

    return sorted(matches)  # Must be sorted for testing


def build_ftn_ctrl_points(code: List[list])-> list:
    """build the nested list of control points"""
    listpath=[]
    root_path=[]
    increment_control_pts = ["if"]
    constant_control_pts = ["case", "else"]
    
    for tokens_line in code:
        if not tokens_line: #empty line
            continue
        if tokens_line[0] in increment_control_pts:
            listpath.append(root_path+[tokens_line[0]])
            root_path.append(tokens_line[0])

        if tokens_line[0] in constant_control_pts:
            listpath.append(root_path+[tokens_line[0]])
            #root_path.append(tokens_line[0])
            
        if tokens_line[0] == "end":
            try:
                if tokens_line[1] in increment_control_pts:
                    try:
                        root_path.pop(-1)
                    except IndexError:
                        msg_err="Error while building controlpoints."
                        logger.warning(msg_err)
                        raise TucanCtrlPtsError(msg_err)
                    
            except IndexError:
                pass
    

    ctrl_points = [[]]
    for path in listpath:
        inserter=ctrl_points
        for _ in path:
            inserter=inserter[-1]
        inserter.extend([path[-1],[]])

    ctrl_points=ctrl_points[0]

        
    return ctrl_points


def compute_complexities_ftn(indents_code: List[int], tokenized_code: List[list]) -> int:
    """Count decision points (if, else if, do, select, etc.)"""
    avg_indent = sum(indents_code)/len(indents_code)/2
    ctrls_pts = build_ftn_ctrl_points(tokenized_code)
    mccabe = compute_cyclomatic_complexity(ctrls_pts)
    possible_paths = compute_possible_paths(ctrls_pts)
    volume, difficulty, effort =  halstead_properties(*halstead_numbers(tokenized_code, PARTS+NESTS+INTRINSICS))
    time_to_code = int(effort/18)

    return round(avg_indent,2),mccabe,possible_paths,volume, difficulty,time_to_code


def compute_cst_ftn(type_: str) -> int:
    """State the structural complexity of a code

    in Short, the average nb. of time we re-read the element.
    It does NOT means it's bad practice
    It just means more read time for the reader to understand the code"""

    cst_ = {
        "program": 0,
        "module": 1,
        "function": 1,
        "subroutine": 2,
        "interface": 4,
        "procedure": 4,
        "type": 6,
    }
    if type_ not in cst_:
        logger.warning(f"Type {type_} not present in conversion list ")
        return 1

    return cst_.get(type_, 1)



from typing import List
from loguru import logger
from tucan.unformat_common import (
    strip_c_comment,
    clean_c_comments,
    read_lines_with_encoding,
)
from tucan.string_utils import tokenize


def scan_ifdef_variables(lines: List[str]) -> list:
    """Detect ifdef variables

    cancels inner variables"""
    out = []
    inner_defined = []

    for line in align_continuation(lines):
        if (
            line.startswith("#ifdef")
            or line.startswith("#elif")
            or line.startswith("#if ")
        ):
            rhs = line.split()[1]

            rhs = rhs.replace("defined(", "")
            rhs = rhs.replace(")", "")

            rhs = rhs.replace("||", " ")
            rhs = rhs.replace("&&", " ")

            out.extend(rhs.split())
        if line.startswith("#define"):
            definition = line.split()
            inner_defined.append(definition[1])

        inner_def = sorted(set(inner_defined))
        global_def = sorted(set(out))
        global_def = [item for item in global_def if item not in inner_defined]

    return global_def, inner_def


def read_definition(line: str) -> str:
    parts = strip_c_comment(line).split()
    out = None
    if len(parts) == 2:
        out = parts[1] + "=True"
    elif len(parts) >= 3:
        out = parts[1] + "=" + " ".join(parts[2:])
    else:
        raise RuntimeError(f"CPP statement - definition not understood :  {line}")
    return out


def remove_ifdef_from_module(
    lines: List[str],
    definitions: List[str],
    verbose: bool = False,
    fortran: bool = False,
) -> List[str]:
    """Cleaned version of a code

    The output must be exactly as long a the input
    """

    uncommented_lines = clean_c_comments(
        align_continuation(lines), fortran=fortran
    )  # we should all C comments, except // for fortran

    out = []
    context = []
    definitions_values = [def_ + "=True" for def_ in definitions]

    local_definitions = []
    for i_line, line in enumerate(uncommented_lines):
        # ifdef_line = line.lstrip()
        if not context:
            included = True
        # ============================== Ifdef logic =============================
        if line.lstrip().startswith("#"):
            out_ = ""
            if line.strip().lstrip("#") == "":
                out.append(out_)
                continue

            tokens = tokenize(line.lstrip().lstrip("#"))

            if tokens[0] in ["ifdef", "ifndef", "elif", "if"]:
                included, context = pragma_start(
                    line, tokens, context, definitions_values + local_definitions
                )
            elif tokens[0] == "else":
                included, context = pragma_else(line, context)
            elif tokens[0] == "endif":
                included, context = pragma_end(line, context, i_line)
            elif tokens[0] in ["define", "undef"]:
                local_definitions = pragma_def_ndef(
                    tokens[0], line, included, local_definitions
                )
            else:  # pragma not treated
                if included:
                    out_ = line
            out.append(out_)

        else:  # Normal code
            if included:
                out.append(line)
                if verbose:
                    logger.success(f"{i_line:04}|{line}")
            else:
                out.append("")
                if verbose:
                    logger.error(f"{i_line:04}|{line}")
    try:
        assert len(out) == len(lines)  # The output length must match exactly the input
    except AssertionError:
        logger.critical(f"Cleaned version {len(out)}/{len(lines)}")
        raise RuntimeError("Ifdef cleaning failed")
    return out


def pragma_start(line, tokens, context, definitions):
    status = str(interpret_ideflogic(line, definitions))
    # increment context
    if tokens[0] == "elif":
        context[-1].append(status)  # append to the sublist of the last element
    else:
        context.append([status])
    included = evaluate_context(context)
    return included, context


def pragma_else(line, context):
    logger.debug(f"Line {line}")
    #  if  elif  else
    #  True False False
    #  False True False
    #  False False True
    # all of the previous if/eli in the context must be false for else to be true
    status = "True"
    for bool_ in context[-1]:
        if bool_ == "True":  # if any of previous is true, status is False.
            status = "False"
    context[-1].append(status)
    included = evaluate_context(context)
    return included, context


def pragma_end(line, context, iline):
    if not context:
        logger.error(f"{iline:04}|{line} : No context found")
        raise RuntimeError("IFdef cleaning failed")
    else:
        context.pop(-1)
    included = evaluate_context(context)
    return included, context


def pragma_def_ndef(token0, line, included, local_definitions):
    if not included:
        return local_definitions

    if token0 == "define":  # variable definitions
        local_definitions.append(read_definition(line))
    if token0 == "undef":  # variable definitions
        def_ = strip_c_comment(line).split()[1]
        local_definitions = [
            item for item in local_definitions if not item.startswith(def_ + "=")
        ]
    return local_definitions


def evaluate_context(context: list) -> bool:
    """Interpret the context to see if next lines will be included or not"""
    if context == []:
        return True
    final_context = [bools_[-1] for bools_ in context]
    if eval(" and ".join(final_context)):
        included = True
    else:
        included = False
    return included


def replace_strings_by_proxies(line: str) -> (str, dict):
    """Extreme measure to handle strings values in Ifdefs..."""
    proxies = {}
    indexes = [i for i, char in enumerate(line) if char == '"']

    if not indexes:
        return line, proxies

    # logger.critical("Replacing strings:"+line)
    last_char = 0
    out_line = ""
    for i in range(0, len(indexes), 2):
        pair = indexes[i : i + 2]
        key = f"#STR{i}#"
        try:
            value = line[pair[0] : pair[1] + 1]
        except IndexError:  # happen is " are not in even number
            break  # well this should never happen but you never know
        proxies[key] = value
        out_line += line[last_char : pair[0] - 1] + " " + key
        last_char = pair[1] + 1

    out_line += line[last_char:]
    return out_line, proxies


def align_continuation(lines: List[str]) -> List[str]:
    """Ensure ligne continuation is understood, but keep line numbers"""

    lines_list = []
    new_line = ""
    skip = 0
    for line in lines:
        new_line += line
        if line.endswith("\\"):
            skip += 1
            new_line = new_line[:-1]  # remove last char
        else:
            lines_list.append(new_line)
            for _ in range(skip):
                lines_list.append("")
            skip = 0
            new_line = ""
    return lines_list


def interpret_ideflogic(line: str, definitions: list) -> bool:
    """Assume an #ifdef-like start:  interpret the content of the line

    NB: no #else or #endif should be ever encountered here (no logic to solve)
    """
    def_dict = {}
    for def_ in definitions:
        key, value = def_.split("=", 1)
        def_dict[key] = value

    re_line = strip_c_comment(line).lstrip().lstrip("#")

    tokens = tokenize(re_line)
    # clean right hand side
    REPLACE = {
        "ifndef": "not",
        "ifdef": "",
        "if": "",
        "elif": "",
        "defined": "",
        "!": "not",
        "||": "or",
        "&&": "and",
    }
    tokens = [REPLACE.get(item, item) for item in tokens]
    re_line = " ".join(tokens)
    # assemble expression as string
    re_line, proxies = replace_strings_by_proxies(re_line)

    expr = ""
    for item in re_line.split():
        expr += " "
        if item in def_dict:
            expr += def_dict[item]
        elif item in proxies:
            expr += proxies[item]
        elif item in [
            "(",
            ")",
            ">",
            "<",  # simple chars operators
            "<=",
            ">=",
            "==",
            "!=",  # double chars operators
            "not",
            "or",
            "and",  # logical operators
        ]:
            expr += item
        elif item.isdigit():
            expr += item
        else:
            expr += "False"
        expr += " "

    try:
        out = eval(expr)
    except (SyntaxError, TypeError, NameError):
        logger.warning(f"Origin :{line}|")
        logger.warning(f"Clean  :{re_line}|")
        logger.warning(f"Expr   :{expr}|")
        # raise RuntimeError("IFDEF expression not understood")
        logger.critical("Fallback to false")
        return False

    return out


def run_ifdef_pkg_analysis(files: dict) -> dict:
    """
    Gather the data associated to the functions and the imports within a file

    Args:
        files (dict): key: short_name , value: absolute paths

    Returns:
        dict: _description_
    """

    ifdef_analysis = {
        "global": [],
        "local": {},
    }

    gvars = []
    for file, path_ in files.items():
        logger.info(f"Detecting iddefs... {path_}")
        lines = read_lines_with_encoding(path_)

        gv_, lv_ = scan_ifdef_variables(lines)

        gvars.extend(gv_)
        ifdef_analysis["local"][file] = lv_

    ifdef_analysis["global"] = sorted(set(gvars))
    logger.success("Analysis completed.")
    return ifdef_analysis


# with open("templates_ifdef.f","r") as fin:
#     lines = fin.read().split("\n")

# vars = scan_ifdef_variables(lines)
# print("Found "+ ", ".join(vars))

# out =remove_ifdef_from_module(lines,["OUIPI1","MOREARG","LVL1"])


# print("\n".join(out))



# """module to guess the language of a code base"""

from loguru import logger
from typing import List


def guess_language(lines: List[str]) -> str:
    """return the language according to lines content
    return None if no clue found"""

    cpp_clues = 0
    ftn_clues = 0
    py_clues = 0
    for line in lines:
        # C/C++ clues
        if line.rstrip().endswith(";"):
            cpp_clues += 1
        for pattern in ["//", "/*", "def ", "#include ", "void"]:
            if line.lstrip().startswith(pattern):
                cpp_clues += 1

        # Python clues
        for pattern in ["'''", '"""', "def ", "import "]:
            if line.lstrip().startswith(pattern):
                py_clues += 1

        # fortran clues
        for pattern in ["function ", "module ", "subroutine ", "implicit "]:
            if line.lstrip().startswith(pattern):
                ftn_clues += 1
        for pattern in ["! ", "     +", "* "]:
            if line.startswith(pattern):
                ftn_clues += 1

    logger.info(f"C {cpp_clues}, Py {py_clues}, FTN {ftn_clues}")
    if cpp_clues > max(py_clues, ftn_clues):
        return "ccpp"
    if ftn_clues > max(py_clues, cpp_clues):
        return "fortran"
    if py_clues > max(ftn_clues, cpp_clues):
        return "python"
    return None



"""Module helper for the CLI of tucan"""
import click
from tucan import __version__ as _ver_
from tucan import __name__ as _name_


def add_version(f):
    """
    Add the version of the tool to the help heading.
    :param f: function to decorate
    :return: decorated function
    """
    doc = f.__doc__
    f.__doc__ = "Package " + _name_ + " v" + _ver_ + "\n\n" + doc
    return f


@click.group()
@add_version
def main():
    r"""
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣤⣄⢀⡀⠀⠀⠀⠀⠀⠀⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣀⣠⣼⣿⣿⣃⠀⠀⠀⠀⠀⠀⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣀⣤⠴⢶⡶⣖⣒⠒⡺⣏⠙⡏⠉⠀⢀⣀⠀⠈⠙⠲⣄⠀⠀⠀⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢠⡾⣫⣤⠀⠀⡰⣿⡇⠀⠁⣽⡆⢷⡖⠛⢉⣭⣉⠳⣄⠀⠈⢧⡀⠀⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢠⣟⠀⠈⠁⠀⠀⠀⠀⠀⠀⠘⣽⣟⠈⣷⡀⣿⣼⢿⠀⢹⠀⠀⠈⢧⠀⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣼⠙⠀⠀⠀⠀⠀⢠⢄⣤⣠⣰⣽⣿⡀⠘⡇⠙⠛⢋⣠⡾⠀⠀⠀⢸⡆⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣿⠓⣸⢼⣟⣛⣛⣿⡿⠻⠛⠻⠏⠁⣉⡽⠋⠉⠉⢉⡞⠁⠀⠀⠀⠀⡇⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠙⠟⠛⠉⠁⠀⠈⠉⠉⠛⠒⡶⠖⠋⠉⠀⠀⠀⣰⠏⠀⠀⠀⠀⠀⠀⡇⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣼⠁⠀⠀⠀⠀⠀⣰⠇⠀⠀⠀⠀⠀⠤⢤⣷⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⡤⠞⠉⠙⢦⠀⠀⠀⠀⠀⠀⣠⢰⠇⠀⠀⠀⠀⠀⢀⡏⢀⡼⠃⠀⠀⠀⠀⠀⢿⡀⠀
                ⠀⠀⠀⠀⠀⠀⢸⡁⠀⠀⠀⠈⢧⡀⠀⠀⠀⠀⠁⣸⠀⠀⠀⠀⠀⠀⣼⠁⡾⠁⠀⠀⠀⠀⠀⠀⠘⡇⠀
                ⠀⠀⠀⠀⠀⠀⠈⢳⡄⠀⠀⠀⠀⢳⡄⠀⠀⠀⠀⡏⠀⠀⠀⠀⠀⢀⡏⢸⠇⠀⠀⠀⠀⠀⠀⠀⠀⢿⠀
                ⠀⠀⡴⠲⠦⣤⣀⡀⢹⡄⠀⠀⠀⠀⠹⡄⠀⠀⠀⡟⢦⡀⠀⢀⣠⠞⠀⢸⡀⠀⠀⠀⠀⠀⠀⠀⠀⢸⠀
                ⠀⠈⠳⠤⣄⣀⠈⠉⠉⠁⠀⠀⠀⡤⠖⠛⡲⣄⠀⡇⠀⠈⠉⠉⠀⠀⠀⠸⣇⠀⠀⠀⠀⠀⠀⠀⠀⢸⡄
                ⠀⠀⠀⣠⣤⣨⣽⡓⠲⢤⣄⡀⠀⠙⢻⠟⣵⣾⣧⣻⡀⠀⠀⠀⠀⠀⠀⠀⠹⣦⡇⠀⠀⠀⠀⠀⠀⢸⡇
                ⠀⠀⡾⣡⣿⡟⣸⢿⣷⡄⠀⠙⣆⠀⠘⠛⠁⠈⢿⠻⣷⡀⠀⢰⡀⠀⠀⠀⠀⠈⣷⠀⢰⠀⢀⠀⠀⢸⠃
                ⠀⠸⠓⠛⠉⠀⠸⣮⣃⡷⠀⠀⠘⣦⠀⠀⠀⠀⠈⠧⣾⠻⣦⡈⢷⣄⠀⠀⠀⢀⣹⣆⣿⡀⢹⠀⠀⣸⠀
                ⠀⠐⠊⠀⠀⠀⠀⠀⠉⠁⠀⠀⠀⠈⢳⡀⠀⠀⠀⠀⠘⣧⠈⠙⣦⣟⢿⡖⠚⠋⠀⠉⠙⣧⣿⡆⢀⡏⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠹⣄⠀⠀⠀⢀⣸⣷⣴⠏⣠⡞⢹⡗⠒⠛⠀⠀⠀⠘⣧⣼⠁⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠙⣦⠀⠰⣏⣡⠾⠋⠻⢯⡀⠀⡇⠀⠀⠀⠀⠀⠀⢹⡃⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠘⣦⠀⢸⣇⡶⠟⠻⣼⠇⠀⡇⠀⠀⠀⠀⠀⠀⠸⡇⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠈⢷⡄⠀⠀⠀⠀⢘⣧⡀⣟⠲⣤⣀⠀⠀⠀⠀⢷⠀⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢻⡄⣠⣶⣿⢿⡏⣿⢹⣄⠀⠉⠛⠲⠶⠶⢾⡆⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢻⣋⣿⣷⣯⠿⠃⠀⠉⢷⣄⣄⠀⠀⠀⠈⡇⠀
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠈⠉⠀⠀⠀⠀⠀⠀⠀⠀⠈⠛⠚⠲⣴⡇⠀

    -----------------------------   TUCAN   -----------------------------

    You are now using the Command line interface of Tucan package,
    a set of tools created at CERFACS (https://cerfacs.fr).
    It is a set of basic helpers around Fortran and Python language

    Checkout anubis and marauder's map packages, two Cerfacs tools
    able to explore respectively the history and geography of codes,
    which both are based upon Tucan.

    """
    pass


@main.command()
@click.argument("filename", type=str, nargs=1)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="dump json data",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose mode",
)
def clean(filename: str, dump: bool, verbose: bool):
    """
    Unformat a fortran of python single file.

    \b
    - Merge multiline statements to one line
    - Split ';' statements
    - Strip comments.
    - Strip blank lines.
    """
    from tucan.unformat_main import unformat_main
    from tucan.tucanlogging import startlog

    startlog(verbose)

    statements = unformat_main(filename, verbose)
    if statements is None:
        print("No statements found...")
        return

    base = filename.split("/")[-1].split(".")[0]
    print(statements)

    statements.dump_code("./" + base + "._rfmt")

    if dump:
        statements.dump_json("./" + base + ".json")


# @main.command()
# @click.argument("path", type=str, nargs=1)
# def package_clean(path):
#     """
#     Unformat a fortran and / or python folder.
#     """

#     import json
#     from loguru import logger

#     from tucan.package_analysis import (
#         rec_travel_through_package,
#         clean_extensions_in_paths,
#         run_unformat,
#     )

#     logger.info("Recursive path gathering ...")
#     paths = rec_travel_through_package(path)
#     logger.info("Cleaning the paths ...")
#     paths = clean_extensions_in_paths(paths)
#     logger.info("Running unformat ...")
#     statements = run_unformat(paths)

#     newfile = "statements_cleaned.json"
#     logger.info(f"Data dumped to {newfile}")
#     with open(newfile, "w") as fout:
#         json.dump(statements, fout, indent=2, sort_keys=True)


@main.command()
@click.argument("filename", type=str, nargs=1)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="dump json data",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose mode",
)
def struct(filename, dump, verbose):
    """
    Extract structure of a fortran or python single file.

    \b
    - Find the nested structures of a code
    - Find the callables in each structure
    - Evaluate sizes, CCN
    """
    import json
    from loguru import logger

    from tucan.struct_main import struct_main
    from tucan.struct_common import struct_summary_str
    from tucan.tucanlogging import startlog
    from tucan.tucanexceptions import TucanError
    startlog(verbose)

    try:
        struct_ = struct_main(filename, verbose)
    except TucanError:
        return
    logger.info("Found following structure:\n" + struct_summary_str(struct_))
    base = filename.split("/")[-1].split(".")[0]
    if dump:
        newfile = base + ".json"
        logger.info(f"Data dumped to {newfile}")
        with open(newfile, "w") as fout:
            json.dump(struct_, fout, indent=2, sort_keys=True)


@main.command()
@click.argument("filename", type=str, nargs=1)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="dump json data",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose mode",
)
def imports(filename, dump, verbose):
    """
    Extract imports of a single file FILENAME
    """
    import json
    from loguru import logger
    from tucan.imports_main import imports_of_file
    from tucan.tucanlogging import startlog

    startlog(verbose)

    imports_, _ = imports_of_file(filename)
    logger.info(json.dumps(imports_, indent=2))
    if dump:
        with open("imports.json", "w") as fout:
            json.dump(imports_, fout, indent=2)


@main.command()
@click.argument("path", type=str, nargs=1)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="dump json data",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose mode",
)
def imports_repo(path, dump, verbose):
    """
    Extract imports of a full repository PATH
    """
    import json
    from loguru import logger

    from tucan.imports_main import imports_of_repository
    from tucan.tucanlogging import startlog

    startlog(verbose)

    imports_, _ = imports_of_repository(path)
    logger.info(json.dumps(imports_, indent=2))
    if dump:
        with open("imports.json", "w") as fout:
            json.dump(imports_, fout, indent=2)


@main.command()
@click.argument("filename", type=str, nargs=1)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="dump json data",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose mode",
)
def ifdef_scan(filename, dump, verbose):
    """
    Extract Ifdef variables of a single file FILENAME.
    """
    import json
    from loguru import logger
    from tucan.clean_ifdef import scan_ifdef_variables
    from tucan.tucanlogging import startlog

    startlog(verbose)

    with open(filename, "r") as fin:
        lines = fin.read().split("\n")

    gv_, lv_ = scan_ifdef_variables(lines)
    gv_s = ", ".join(gv_)
    lv_s = ", ".join(lv_)

    logger.info(f"Global ifdef variables : {gv_s}")
    if lv_:
        logger.info(f"Found local ifdef variables : {lv_s}")
    else:
        logger.info("No local ifdef variables")

    if dump:
        base = filename.split("/")[-1].split(".")[0]
        newfile = base + "_ifdefs.json"
        logger.info(f"Data dumped to {newfile}")
        out = {"global": gv_, "local": lv_}
        with open(newfile, "w") as fout:
            json.dump(out, fout, indent=2, sort_keys=True)


@main.command()
@click.argument("filename", type=str, nargs=1)
@click.option(
    "-def",
    "--definitions",
    type=str,
    default=None,
    # multiple=True,
    help="definitions to resolve ifdefs. Comma separated ',', no spaces : -v ARG1,ARG2",
)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="dump json data",
)
@click.option(
    "-f",
    "--fortran",
    is_flag=True,
    help=" fortran mode (avoid  stripping //)",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose execution",
)
def ifdef_clean(filename, definitions, dump, fortran, verbose):
    """
    Show a file FILENAME with idefs resolved
    """
    from loguru import logger
    from tucan.clean_ifdef import remove_ifdef_from_module
    from tucan.tucanlogging import startlog

    startlog(verbose)

    if definitions is None:
        definitions = []
    else:
        definitions = definitions.split(",")

    with open(filename, "r") as fin:
        lines = fin.read().split("\n")
    lines = remove_ifdef_from_module(
        lines, definitions, verbose=verbose, fortran=fortran
    )
    logger.success("Ifdefs resolved:")
    for line in lines:
        print(line)
    if dump:
        newfile = filename + "_ifdef_resolved"
        lines.append("# the ifdefs were resolved by tucan")
        v_s = ", ".join(definitions)
        lines.append(f"# IFdef Variables: {definitions}")
        with open(newfile, "w") as fout:
            fout.write("\n".join(lines))


@main.command()
@click.argument("path", type=str, nargs=1)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="dump json data",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose execution",
)
def ifdef_scan_repo(path, dump, verbose):
    """
    Extract Ifdef variables of a full repository PATH
    """
    import json, os
    from loguru import logger
    from tucan.clean_ifdef import run_ifdef_pkg_analysis
    from tucan.package_analysis import (
        rec_travel_through_package,
        get_package_files,
    )
    from tucan.tucanlogging import startlog

    startlog(verbose)

    logger.info("Recursive path gathering ...")
    paths = rec_travel_through_package(path)
    logger.info("Cleaning the paths ...")
    files = get_package_files(paths, os.getcwd())

    out = run_ifdef_pkg_analysis(files)
    gv_s = ", ".join(out["global"])
    logger.info(f"Global ifdef variables : {gv_s}")
    for file, lv_ in out["local"].items():
        if lv_:
            lv_s = ", ".join(lv_)
            logger.info(f"Local to {file} : {lv_s}")

    if dump:
        newfile = path + "package_ifdefs.json"
        logger.info(f"Data dumped to {newfile}")
        with open(newfile, "w") as fout:
            json.dump(out, fout, indent=2, sort_keys=True)


@main.command()
@click.argument("path", type=str, nargs=1)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="verbose execution",
)
def struct_repo(path, verbose):
    """
    Extract struct of a repository PATH
    """
    from loguru import logger
    import json, os
    from tucan.travel_in_package import (
        rec_travel_through_package,
        get_package_files,
    )
    from tucan.package_analysis import (
        run_full_analysis,
    )
    from tucan.tucanlogging import startlog

    startlog(verbose)

    logger.info("Recursive path gathering ...")
    paths = rec_travel_through_package(path)
    logger.info("Get files ...")
    files = get_package_files(paths, os.getcwd())
    logger.info("Running struct ...")
    full_analysis = run_full_analysis(files)

    newfile = "package_analysis.json"
    logger.info(f"Data dumped to {newfile}")
    with open(newfile, "w") as fout:
        json.dump(full_analysis, fout, indent=2, sort_keys=True)



"""
Tucan module to get the imports of a python file
"""

from typing import List

# from tucan.unformat_common import Statements
from tucan.string_utils import tokenize

# def imports_py(statements: Statements) -> dict:
#     """
#     Main function to search the imports in a python file.

#     Args:
#         statements (Statements): List of unformatted code statements along with line number ranges.

#     Returns:
#         dict: Dict with the module or function name associated with its various imports.
#     """
#     imports_dict = python_imports(statements)

#     return imports_dict


# def python_imports(stmts: Statements) -> dict:
#     """
#     Parse the python code to check for the imports and update the imports database.

#     Args:
#         stmts (Statements): List of unformatted code statements along with line number ranges.

#     Returns:
#         imports_dict (dict): Dict with the module or function name associated with its various imports.
#     """
#     imports_dict = {}
#     for line in stmts.stmt:
#         if "import " not in line:
#             pass

#         # Parsing "as" imports (ex: numpy as np) for later identification
#         if line.strip().startswith("import "):
#             module_from = line.split("import ")[-1].strip()
#             if " as " in line:
#                 alias = line.split(" as ")[-1].strip()
#                 module_from = module_from.split(" as ")[0].strip()
#                 imports_dict[module_from] = {"alias": alias, "explicit": []}
#             else:
#                 module_from = module_from.split(",")
#                 for mod_from in module_from:
#                     imports_dict[mod_from.strip()] = {"explicit": []}

#         # Parsing "from XXX import XXX", i.e. specific imports
#         if line.strip().startswith("from ") and "import" in line.strip():
#             module_from = line.split("import ")[0].split("from ")[-1].strip()
#             func_imported = line.split("import ")[1:]
#             # Handle the case from XXXX import XXXX as XXXX
#             if "as " in func_imported[0]:
#                 mod_and_alias = func_imported[0].split("as ")
#                 func_imported = {
#                     "alias": mod_and_alias[1].strip(),
#                     "explicit": [mod_and_alias[0].strip()],
#                 }  # Name of function is saved not the alias yet

#             # Handle imports on multiple lines
#             elif func_imported[0].startswith("("):
#                 func_imported = func_imported[0].split("(")[-1]
#                 func_imported = func_imported.split(")")[0]
#                 func_imported = {
#                     "explicit": [func_.strip() for func_ in func_imported.split(",")]
#                 }

#             else:
#                 func_imported = func_imported[0].split(",")
#                 func_imported = {"explicit": [func_.strip() for func_ in func_imported]}

#             imports_dict[module_from] = func_imported

#     return imports_dict


def _by_import(tokens):
    """Resolve a Python direct import line"""
    # namespaces
    if "as" in tokens:

        if len(tokens) != 4:
            line = " ".join(tokens)
            raise RuntimeError(f"Unexpected situation: {line}")
        init_ref = tokens[1]
        ref = tokens[1].split(".")[0]
        #return {tokens[3]: f"{ref}.{tokens[1]}"}
        return {tokens[3]: f"__py__@{tokens[1]}"}

    # multiple imports
    if "," in tokens:
        out = {}
        for token in tokens:
            if token not in ["import", ",", "(", ")"]:
                #out[token] = token
                out[token] = f"__py__@{token}"
        return out
    # default
    ref = tokens[1].split(".")[0]
        
    return {tokens[1]: f"__py__@{tokens[1]}"}


def _by_from(tokens):
    """Resolve a Python from import line"""
    # namespaces

    retokens = [tok for tok in tokens if tok not in [",", "(", ")"]]

    ref = retokens[1].lstrip(".")

    out = {}
    if "as" in retokens:
        idx = retokens.index("as")
        for i in range(0, idx - 3):
            name = retokens[3 + i]
            alias = retokens[idx + 1 + i]
            out[alias] = f"__py__@{ref}.{name}"
    else:
        if retokens[3] == "*":
            return {f"_{ref}": f"__py__@{ref}.*"}
          
        for i in range(0, len(retokens) - 3):
            name = retokens[3 + i]
            out[name] = f"__py__@{ref}.{name}"

    return out


def get_imports_py(lines: List[str]) -> dict:
    """Read the imports information from python lines"""
    out = {}
    for line in lines:
        if line.lstrip().startswith(("from", "import")):
            tokens = tokenize(line)
            if tokens[0] == "import":
                out.update(_by_import(tokens))
            if tokens[0] == "from":
                out.update(_by_from(tokens))

    return out



import re
from typing import Tuple, List
from loguru import logger
from tucan.unformat_common import Statements
from tucan.struct_common import (
    new_buffer_item,
    new_stack_item,
    struct_from_stack,
    struct_augment,
    list2pathref,
)
from tucan.string_utils import find_words_before_left_parenthesis_noregexp

from tucan.kw_lang import KEYWORDS_PY
from tucan.complexities import (
    halstead_numbers,
    halstead_properties,
    compute_cyclomatic_complexity,
    compute_possible_paths,
)
from  tucan.tucanexceptions import TucanCtrlPtsError

def extract_struct_py(statements: Statements, verbose: bool) -> dict:
    """Main calls to build structure form statements

    statements is the output of tucan.unformat_py.unformat_py
    """
    # TODO: verbose not supported for python
    clean_code = statements.to_code()
    all_structs = _extract_on_cleaned_py(statements)
    all_structs = struct_augment(
        all_structs,
        clean_code,
        find_callables_py,
        compute_complexities_py,
        compute_cst_py,
    )
    for part, data in all_structs.items():
        if data["type"] == "class":
            data["parents"].extend(find_class_inheritance(data["linestart"]))
    return all_structs


def _extract_on_cleaned_py(stmts: Statements) -> dict:
    """Extract structure from cleaned statements."""
    buffer = []
    stack = []
    path = []
    last_indent = 0
    last_line = None
    stat_idx = -1
    for line, (line_idx1, line_idx2) in zip(stmts.stmt, stmts.lines):
        stat_idx += 1
        indent = int(len(re.findall(r"\s{4}|\s\t+", line)))

        # Find correctly the indentation without accessing spaces in strings
        indent = re.findall(r"^\s+", line)
        if indent:
            indent = int(len(indent[0]) / 4)
        else:
            indent = 0

        # Evaluate indentation level for path and buffer
        if indent > last_indent:
            if indent > last_indent + 1:
                logger.warning(
                    f"Multiple indent at {path} for '({line_idx1}-{line_idx1}){line}' with last indent : {last_indent} and current : {indent}"
                )
                indent = last_indent + 1
            type_, name = parse_name_py(last_line)

            path.append(name)
            buffer.append(
                new_buffer_item(
                    type_=type_,
                    path=path,
                    name=name,
                    first_line=last_line,
                    line_idx=last_idx,
                    statement_idx=stat_idx,
                )
            )
            last_line = line
            last_idx = line_idx2
            last_indent = indent
            continue

        elif indent < last_indent:
            for _ in range(last_indent - indent):
                last_buff = buffer[-1]
                stack.append(new_stack_item(last_buff, last_idx, stat_idx, line))
                path.pop(-1)
                buffer.pop(-1)
            last_line = line
            last_indent = indent
            last_idx = line_idx1
            continue

        last_line = line
        last_indent = indent
        last_idx = line_idx1

    struct = struct_from_stack(
        stack, main_types=["def", "class"], skip_types=["if", "for"]
    )

    struct = explicit_methods(struct)

    return struct


def explicit_methods(struct: dict):
    """prepend names of methods by their classes name"""
    for part, data in struct.items():
        path = data["path"]
        if len(path) > 1:
            father_part = list2pathref(path[:-1])
            if father_part in struct:
                if struct[father_part]["type"] == "class":
                    data["name"] = struct[father_part]["name"] + "." + data["name"]
                    data["type"] = "method"
    return struct


def parse_name_py(line: str) -> Tuple[str, str]:
    """expect a lowercase stripped line
    takes the second word as the name
    """
    type_ = line.strip().split()[0]

    try:
        name = line.strip().replace("(", " ").split()[1].replace(":", " ").split()[0]
    except IndexError:
        name = "dummy"

    return type_, name


##### Main structs


def find_callables_py(tokenized_code: List[list]) -> list:
    """Find callables in python"""
    candidates = []
    for tokens in tokenized_code:
        # if not line.strip().startswith("def") and not line.strip().startswith("class"):
        candidates.extend(find_words_before_left_parenthesis_noregexp(tokens))
    matches = [cand.strip() for cand in set(candidates) if cand not in KEYWORDS_PY]

    return sorted(matches)  # Must be sorted for testing


def find_annotations_from_args_py(line: str) -> dict:
    """Find annotation in arguments"""
    id1 = line.find("(")
    id2 = line.rfind(")")
    out = {}
    for arg in line[id1 + 1 : id2].split(","):
        if ":" in arg:
            arg, type_ = arg.split("=")[0].split(":")
            out[arg] = type_
    return out


def find_class_inheritance(line: str) -> list:
    """Find the potential inheritance in a class"""
    id1 = line.find("(")
    id2 = line.rfind(")")
    if id1 == -1:
        return []

    out = line[id1 + 1 : id2].strip()
    if "," in out:
        return [item.strip() for item in out.split(",")]
    elif len(out) == 0:
        return []
    else:
        return [out]



def build_py_ctrl_points(indents: List[int],code: List[list])-> list:
    """build the nested list of control points for python"""
    
    if code == []:
        return []
    
    increment_control_pts = ["if",]
    constant_control_pts = ["elif", "else"]
    
    listpath=[]
    root_path=[]

    # get exhaustive list of levels
    last_lvl=indents[0]
    last_token_line=code[0]
    for current_lvl,tokens_line in zip(indents[1:],code[1:]):
        try: 
            anchor = last_token_line[0]
        except IndexError:
            anchor = None

        if current_lvl>last_lvl:
            root_path.append(anchor)
            listpath.append(root_path.copy())
            
        if current_lvl<last_lvl:
            while current_lvl<last_lvl:
                try:
                    root_path.pop(-1)
                except IndexError:
                    msg_err="Error while building controlpoints. Check the indentation maybe?"
                    logger.warning(msg_err)
                    raise TucanCtrlPtsError(msg_err)
                last_lvl-=1
        
        last_lvl=current_lvl
        last_token_line=tokens_line

    # remove non-control-points levels      
    for i,path in enumerate(listpath):
        listpath[i] =  [item for item in path if item in increment_control_pts+constant_control_pts]

    # move to a nested list
    ctrl_points = [[]]
    for path in listpath:
        if path == []:
            continue
        inserter=ctrl_points
        for _ in path:
            inserter=inserter[-1]
        inserter.extend([path[-1],[]])
    ctrl_points=ctrl_points[0]
    return ctrl_points


def compute_complexities_py(indents_code: List[int], tokenized_code: List[list]) -> int:
    """Count decision points (if, else if, do, select, etc.)"""
    indents_code=[ind/4 for ind in indents_code]
    avg_indent = sum(indents_code)/len(indents_code)
    ctrls_pts = build_py_ctrl_points(indents_code,tokenized_code)
    mccabe = compute_cyclomatic_complexity(ctrls_pts)
    possible_paths = compute_possible_paths(ctrls_pts)
    volume, difficulty, effort =  halstead_properties(*halstead_numbers(tokenized_code, KEYWORDS_PY))
    time_to_code = int(effort/18)

    return round(avg_indent,2),mccabe,possible_paths,volume, difficulty,time_to_code


def compute_cst_py(type_: str) -> int:
    """State the structural complexity of a code

    in Short, the average nb. of time we re-read the element.
    It does NOT means it's bad practice
    It just means more read time for the reader to understand the code"""

    cst_ = {"def": 1, "class": 4, "method": 2}
    if type_ not in cst_:
        logger.warning(f"Type {type_} not present in conversion list ")
        return 1

    return cst_.get(type_, 1)



"""
Tucan module to get the imports of a python file
"""
from typing import List
from tucan.string_utils import tokenize
from loguru import logger

def by_cpp_include(tokens):
    """Resolve a CPP include import, also used in Fortran"""
    # namespaces
    if tokens[2] == "<":
        ref = tokens[3]
    else:
        ref = tokens[2].strip("'").strip('"')

    ref = ref.replace("../","")
    return {f"_{ref}": ref+"@*"}


def get_imports_ccpp(lines: List[str]) -> dict:
    """Read the imports information from C/C++ lines"""
    out = {}
    for line in lines:
        if line.lstrip().startswith("#"):
            line = line.replace("<","'").replace(">","'")
            tokens = tokenize(line)
            if tokens[1] in ["include", "INCLUDE"]:
                out.update(by_cpp_include(tokens))
    return out



from typing import List, Tuple
from loguru import logger
from tucan.kw_lang import KEYWORDS_FTN

from tucan.unformat_common import (
    Statements,
    new_stmts,
    remove_strings,
    clean_blanks,
    clean_inline_comments,
    clean_pure_comments,
    align_multiline_blocks,
    split_multi_statement_lines,
    align_end_continuations,
)
from tucan.string_utils import eat_spaces, get_indent, tokenize


def align_start_continuations(stmts: Statements) -> Statements:
    """Align continuation with amprsand

    For example

     FORMAT("Lorem ipsum
    & sic hamet")

    """
    new_stmt = []
    new_lines = []
    last_line = ""
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if line.lstrip().startswith("&"):
            new_stmt[-1] = last_line.rstrip("&") + " " + line.strip()[1:]
            new_lines[-1][1] = lend
        elif line.lstrip().startswith("$"):
            new_stmt[-1] = last_line.rstrip("$") + " " + line.strip()[1:]
            new_lines[-1][1] = lend
        elif line.startswith("     +"):
            new_stmt[-1] = last_line + " " + line.strip()[1:]
            new_lines[-1][1] = lend
        elif line.startswith("     ."):
            new_stmt[-1] = last_line + " " + line.strip()[1:]
            new_lines[-1][1] = lend
        else:
            new_stmt.append(line)
            new_lines.append([lstart, lend])

        last_line = new_stmt[-1]
    return Statements(new_stmt, new_lines)


def ftn_align_labelled_continuations(stmts: Statements) -> Statements:
    """Align labelled continuations , very old style
       15 FORMAT(
         1/' TRANFT:  Transport property fitting,',
         2/'           CHEMKIN-II Version ',A,', August 1994',
    C*****precision > double
         3/'           DOUBLE PRECISION')

         becomes
        15 FORMAT(' TRANFT:  Transport property fitting,',' CHEMKIN-II Version ',A,', August 1994','DOUBLE PRECISION')

    """
    new_stmt = []
    new_lines = []
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if line[0:5] == "     " and (line[5] in "0123456789*"):
            new_stmt[-1] += line[7:]
            new_lines[-1][1] = lend
        else:
            new_stmt.append(line)
            new_lines.append([lstart, lend])

    return Statements(new_stmt, new_lines)


def ftn_make_oneliners_conditionals_multilines(stmts: Statements) -> Statements:
    """_summary_

    Args:
        stmts (Statements): _description_

    Returns:
        Statements: _description_
    """
    new_stmt = []
    new_lines = []
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if (
            line.strip().startswith("if ")
            and " then" not in line
            and not line.strip().startswith("if =") # extreme case of a variable named if
        ):
            # logger.critical(line)
            stmts_, lines_  = split_if_oneliner( line, lstart, lend)
            new_stmt.extend(stmts_)
            new_lines.extend(lines_)

        elif line.strip().startswith("where "):
            stmts_, lines_  = split_where_oneliner( line, lstart, lend)
            new_stmt.extend(stmts_)
            new_lines.extend(lines_)

        else:
            new_stmt.append(line)
            new_lines.append([lstart, lend])
    return Statements(new_stmt, new_lines)


def split_if_oneliner(line, lstart, lend):
    head, tail = split_after_1st_parenthesis_group(line)
    new_stmt =  []
    new_lines = []
    
    if tail == "":
        logger.warning(f"Cannot split properly line :{line} ")
    
    indent = get_indent(line)
    
    new_stmt.append(head + " then")
    new_lines.append([lstart, lend])
    
    tail = indent + "    " + tail
    if tail.lstrip().startswith("where"): # nested if and where
        stmts_, lines_  = split_where_oneliner( tail, lstart, lend)
        new_stmt.extend(stmts_)
        new_lines.extend(lines_)
    else:
        new_stmt.append(tail)
        new_lines.append([lstart, lend])
    
    new_stmt.append(indent + "end if")
    new_lines.append([lstart, lend])
    return new_stmt, new_lines


def split_where_oneliner(line, lstart, lend):
    head_, tail = split_after_1st_parenthesis_group(line)
    
    new_stmt =  []
    new_lines = []

    if tail == "": # already splitted
        new_stmt.append(head_)
        new_lines.append([lstart, lend])
        return new_stmt, new_lines 
    
    indent = get_indent(line)
    
    
    new_stmt.append(head_)
    new_lines.append([lstart, lend])
    
    tail = indent + "    " + tail
    new_stmt.append(tail)
    new_lines.append([lstart, lend])
    
    new_stmt.append(indent + "end where")
    new_lines.append([lstart, lend])
    return new_stmt, new_lines 




def split_after_1st_parenthesis_group(line: str) -> Tuple[str, str]:
    """
    Splits a line into two parts: the first complete parentheses block and the remainder of the line.

    Args:
        line (str): A string that may contain parentheses.

    Returns:
        Tuple[str, str]: A tuple where the first element is the substring up to and including the first complete
                         parentheses block, and the second element is the remainder of the string after this block.
                         If there is no closing parenthesis, the first element will contain everything from the line,
                         and the second element will be an empty string.
    """
    path = []        # Stack to track open parentheses
    head=""
    tail=""
    for idx, char in enumerate(line):
        head += char
        if char == "(":
            path.append(char)
        elif char == ")":
            if path:
                path.pop()
            if not path:  # If stack is empty, we closed all opened parentheses
                tail = line[idx + 1:].lstrip()
                break
    return head,tail


def ftn_remove_space_in_front_of_variables(lines: List[str]) -> List[str]:
    """
    Detect keywords used as variables, and make sure it is without space:

    TYPE = 300
    becomes
    TYPE= 300

    This way, a true keyword is always followed by a space, and will be detected,
    but false keywords wont
    """
    new_stmt = []
    for line in lines:
        stmt = line
        try:
            item0 = line.split()[0]
            item1 = line.split()[1]
        except IndexError:
            item1 = None

        if item1 == "=" and item0 in KEYWORDS_FTN:
            stmt = line.replace(line.split()[0] + " =", "__" + line.split()[0] + "__=")
            logger.warning(f"Fortran Keywords {item0} used as a variable in the code.")

        new_stmt.append(stmt)

    return new_stmt


def ftn_clean_labelled_loops_oldftn(stmts: Statements) -> Statements:
    """Clean DO with labels instead of END DO

    Example in tranfit
    DO 600 K = 1, KK
        IF (NLIN(K) .EQ. NCHECK) THEN
            DO 750 J = K, KK
                IF (NLIN(J) .EQ. NCHECK) WRITE (LOUT, 8010) KSYM(J)
            750  CONTINUE
            KERR = .TRUE.
        ENDIF
    600 CONTINUE

    to

    DO 600 K = 1, KK
        IF (NLIN(K) .EQ. NCHECK) THEN
            DO 750 J = K, KK
                IF (NLIN(J) .EQ. NCHECK) WRITE (LOUT, 8010) KSYM(J)
            END DO !750
            KERR = .TRUE.
        ENDIF
    END DO !600
    """
    new_stmt = []
    new_lines = []
    do_loop_lbl = []
    do_loop_indent = []
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        stmt = line

        try:
            item0 = line.split()[0]
            item1 = line.split()[1]
        except IndexError:
            new_stmt.append(stmt)
            new_lines.append([lstart, lend])
            continue

        if item0 == "do":
            tokens = tokenize(line)
            item1=tokens[1] 
            if item1.isdigit():
                value = int(item1)
                do_loop_lbl.append(value)
                do_loop_indent.append(get_indent(line))

        if item0.isdigit():
            value = int(item0)
            if value not in do_loop_lbl:
                new_stmt.append(stmt)
                new_lines.append([lstart, lend])
            else:
                while do_loop_lbl and int(item0) == do_loop_lbl[-1]:
                    stmt = do_loop_indent[-1] + f"end do ! {do_loop_lbl[-1]}"
                    do_loop_indent.pop(-1)
                    do_loop_lbl.pop(-1)
                    new_stmt.append(stmt)
                    new_lines.append([lstart, lend])

        else:
            new_stmt.append(stmt)
            new_lines.append([lstart, lend])

    return Statements(new_stmt, new_lines)


def ftn_suppress_goto_references(lines: List[str]) -> List[str]:
    """Remove the GOTO references , put them at the end of the line"""
    new_stmt = []
    for line in lines:
        stmt = line
        for keyword in KEYWORDS_FTN:
            if keyword in line.split() and line.split()[0].isdigit():
                stmt = (
                    line.replace(line.split()[0], len(line.split()[0]) * " ")
                    + "  ! ("
                    + line.split()[0]
                    + ")"
                )
        new_stmt.append(stmt)
    return new_stmt


def ftn_clean_labelled_loops_newftn(lines: List[str]) -> List[str]:
    """Move modern labels at the end of the line

    We keep label to show the name in struct
    """
    new_stmt = []

    def _move_labels(line: str, pattern: str = "do", sep: str = ":") -> str:
        """
        eigenvalues: block

        becomes block ! NESTING_LABEL# eigenvalues
        """
        if pattern not in line or sep not in line:
            return line

        rhs = line.split(sep)[1].rstrip()
        rhs1 = ""
        for char in rhs:
            if char in " (":
                break
            rhs1 += char

        if rhs1 == pattern:
            label = line.split(sep)[0].strip()
            index = line.find(sep)
            line = (
                get_indent(line)
                + line[index + 1 :].rstrip()
                + " ! NESTING_LABEL# "
                + label
            )
        return line

    for line in lines:
        line = _move_labels(line, "block", ":")
        line = _move_labels(line, "do", ":")
        line = _move_labels(line, "if", ":")
        line = _move_labels(line, "entry", ":")
        line = _move_labels(line, "select_case", ":")
        new_stmt.append(line)

    return new_stmt


def ftn_clean_operators(lines: List[str]) -> List[str]:
    new_stmt = []
    for line in lines:
        line = line.replace("::", " :: ")
        line = line.replace("==", ".eq.")
        line = line.replace("/=", ".ne.")
        line = line.replace(">=", ".ge.")
        line = line.replace("<=", ".le.")
        # line =line.replace("=", " = ")
        new_stmt.append(line)

    return new_stmt


def ftn_clean_intrinsics_declarations(lines: List[str]) -> List[str]:
    """clean oddities in fortran declarations"""
    new_stmt = []

    def _remove_decorators(line: str, keyword: str) -> str:
        """usual cleanup before fortran structures:
        - check ends
        - put decorators at the back
        """
        if keyword not in line:
            return line

        if line.lstrip().startswith("end"):
            line.replace("end" + keyword, "end " + keyword)
            return line

        line_no_par = (
            line.replace("(", " ").replace(")", " ") + " "
        )  # Here we add " " to make sure a keyword last on the line still trigger (" "+keyword+" ") for the replacement
        idx = line_no_par.find(" " + keyword + " ")
        if idx == -1:
            return line

        line = (
            get_indent(line)
            + line[idx + 1 :]
            + " ! "
            + keyword.upper()
            + "# "
            + line[: idx + 1].lstrip()
        )
        return line

    for line in lines:
        # merge multiple words
        line = line.replace("double precision", "double_precision ")
        line = line.replace("module procedure", "module_procedure ")
        line = line.replace("selectcase", "select_case ")
        line = line.replace("select case", "select_case ")
        line = line.replace("select type", "select_type ")
        line = line.replace("where(", "where (")

        line = _remove_decorators(line, "function")
        line = _remove_decorators(line, "subroutine")
        line = _remove_decorators(line, "interface")

        line = line.replace("enum,", "enum ,")

        line = line.replace("enddo", "end do ")
        line = line.replace("endsubroutine", "end subroutine ")
        line = line.replace(
            "endif", "end if "
        )  # Important to add supp. spacing , incase of a following elmt.
        line = line.replace("endwhere", "end where ")
        line = line.replace(
            "endmodule", "end module "
        )  # Important to add supp. spacing , incase of a following elmt.

        # slpit left parenthesis - and right ...
        for keyword in [
            "function",
            "subroutine",
            "interface",
            "if",
            "do",
            "type",
            "associate",
        ]:
            line = line.replace(f"{keyword}(", f"{keyword} (")

        line = line.replace(")then", ") then ")

        new_stmt.append(line)

    return new_stmt


def ftn_clean_type_keywords(lines: List[str]) -> List[str]:
    """Replace  type( declarations by _type("""

    def _clean_type(line: str) -> str:
        tokens = tokenize(line)
        if tokens[1] in ["is"]:
            tokens[0] = "__type_is__"
            del tokens[1]
            line = get_indent(line) + " ".join(tokens)
        elif tokens[1] in ["("]:
            tokens[0] = "__type_from__"
            line = get_indent(line) + " ".join(tokens)
        elif tokens[1] in [","]:
            line = get_indent(line) + " ".join(tokens)
        else:
            pass
        return line

        # if line.replace(" ","")[4]=="(":         #for type(bcdhjzks)
        #     return line.replace("type","_type",1)
        # elif line.replace(" ","")[4:6]=="is":    #for select type ; type is
        #     return line.replace("type is","#typeis",1)
        # elif line.replace(" ","")[4] in " azertyuiopqsdfghjklmwxcvbn":        #well formed keyword
        #     return line
        # elif line.replace(" ","")[4]==",":
        #     return line.replace("type","type ",1)   #for  type, extend (need the space to identify the end of keyword)
        # else:
        #     return line

    new_stmt = []
    for line in lines:
        if line.lstrip().startswith("type"):
            line = _clean_type(line)

        new_stmt.append(line)
    return new_stmt


def split_multi_declaration_lines(stmts: Statements) -> Statements:
    """Split statements on ;"""
    new_stmt = []
    new_lines = []
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if "," in line and line.lstrip().startswith(("procedure", "module_procedure")):
            indent = get_indent(line)
            tokens = tokenize(line)
            try:
                min_idx = tokens.index("::")
            except ValueError:  # no "," after "::"
                min_idx = 0

            try:
                idx = tokens.index(",", min_idx)
            except ValueError:  # no "," after "::"
                new_stmt.append(line)
                new_lines.append([lstart, lend])
                continue

            new_stmt.append(indent + " ".join(tokens[:idx]))  # case "," after "::"
            new_lines.append([lstart, lend])
            for token in tokens[idx:]:
                if token != ",":
                    new_stmt.append(indent + " ".join(tokens[: idx - 1]) + " " + token)
                    new_lines.append([lstart, lend])
        else:
            new_stmt.append(line)
            new_lines.append([lstart, lend])

    return Statements(new_stmt, new_lines)


def align_multiline_blocks_ftn(
    stmts: Statements
) -> Statements:
    merge_ = []
    parenthesis_lvl = 0
    ml_in = None
    ml_lvl = None
    for i, line in enumerate(stmts.stmt):
        for j,char in enumerate(line):
            if char == "(":
                parenthesis_lvl +=1
               
                try:
                    if line[j+1] == "/":
                        ml_in=i
                        ml_lvl=parenthesis_lvl
                        
                        
                except IndexError:
                    pass
            elif char == ")":
                if parenthesis_lvl == ml_lvl:
                    merge_.append((ml_in,i))
                    ml_in = None
                    ml_lvl = None
                    
                parenthesis_lvl -=1
                
                
    new_stmt = stmts.stmt
    new_lines = stmts.lines

    def replace_slice_(lst, start, end, new_item):
        return lst[:start] + [new_item] + lst[end + 1 :]

    for i_in_, i_out_ in merge_:
        tmp_line = [stmts.lines[i_in_][0], stmts.lines[i_out_][1]]

        tmp_stmt = get_indent(stmts.stmt[i_in_]) + " ".join(
            [item.lstrip() for item in stmts.stmt[i_in_ : i_out_ + 1]]
        )

        new_stmt = replace_slice_(new_stmt, i_in_, i_out_, tmp_stmt)
        new_lines = replace_slice_(new_lines, i_in_, i_out_, tmp_line)

    return Statements(new_stmt, new_lines)

def unformat_ftn(code: List[str]) -> Statements:
    """
    Unformat Fortran code by stripping comments and moving leading '&' characters.

    Args:
        code (List[str]): List of Fortran code lines.

    Returns:
        List[Tuple[str, Tuple[int, int]]]: List of unformatted code statements along with line number ranges.
    """
    stmts = new_stmts(code)
    #return stmts
    stmts = clean_pure_comments(stmts, "c")
    stmts = clean_pure_comments(stmts, "C")
    stmts = clean_pure_comments(stmts, "*")

    stmts.stmt = eat_spaces(stmts.stmt)
    stmts.stmt = remove_strings(stmts.stmt, '"')
    stmts.stmt = remove_strings(stmts.stmt, "'")

    stmts = clean_inline_comments(stmts, "!")
    stmts = clean_blanks(stmts)

    stmts = align_end_continuations(stmts, "&")
    stmts = align_start_continuations(stmts)
    stmts = ftn_align_labelled_continuations(stmts)
    stmts = align_multiline_blocks_ftn(stmts)

    stmts.stmt = ftn_clean_operators(stmts.stmt)
    stmts = split_multi_statement_lines(stmts)

    # stmts.stmt = ftn_clean_type_declarations(stmts.stmt) # not sure we need this now
    stmts.stmt = ftn_clean_type_keywords(stmts.stmt)
    stmts.stmt = ftn_clean_intrinsics_declarations(stmts.stmt)
    stmts = ftn_clean_labelled_loops_oldftn(stmts)
    stmts.stmt = ftn_clean_labelled_loops_newftn(stmts.stmt)

    stmts = split_multi_declaration_lines(stmts)

    stmts.stmt = ftn_suppress_goto_references(
        stmts.stmt
    )  # !!After ftn_clean_labelled_loops_oldftn, or you cannot end loops
    stmts = ftn_make_oneliners_conditionals_multilines(stmts)
    stmts.stmt = ftn_remove_space_in_front_of_variables(stmts.stmt)

    return stmts



KEYWORDS_C = [
    "alignas",
    "alignof",
    "auto",
    "bool",
    "break",
    "case",
    "char",
    "const",
    "constexpr",
    "continue",
    "default",
    "do",
    "double",
    "else",
    "enum",
    "extern",
    "false",
    "float",
    "for",
    "goto",
    "if",
    "inline",
    "int",
    "long",
    "nullptr",
    "register",
    "restrict",
    "return",
    "short",
    "signed",
    "sizeof",
    "static",
    "static_assert",
    "struct",
    "switch",
    "thread_local",
    "true",
    "typedef",
    "typeof",
    "typeof_unqual",
    "union",
    "unsigned",
    "void",
    "volatile",
    "while",
    "_Alignas",
    "_Alignof",
    "_Atomic",
    "_BitInt",
    "_Bool",
    "_Complex",
    "_Decimal128",
    "_Decimal32",
    "_Decimal64",
    "_Generic",
    "_Imaginary",
    "_Noreturn",
    "_Static_assert",
    "_Thread_local",
    "printf",
    "snprintf",
    "strlen",
    "strcmp",
    "strcpy",
    "strncmp",
    "max",
    "min",
]


KEYWORDS_PY = [
    "abs",
    "aiter",
    "all",
    "anext",
    "any",
    "ascii",
    "bin",
    "bool",
    "breakpoint",
    "bytearray",
    "bytes",
    "callable",
    "chr",
    "classmethod",
    "compile",
    "complex",
    "delattr",
    "dict",
    "dir",
    "divmod",
    "enumerate",
    "eval",
    "exec",
    "filter",
    "float",
    "format",
    "frozenset",
    "getattr",
    "globals",
    "hasattr",
    "hash",
    "help",
    "hex",
    "id",
    "input",
    "int",
    "isinstance",
    "issubclass",
    "iter",
    "len",
    "list",
    "locals",
    "map",
    "max",
    "memoryview",
    "min",
    "next",
    "object",
    "oct",
    "open",
    "ord",
    "pow",
    "print",
    "property",
    "range",
    "repr",
    "reversed",
    "round",
    "set",
    "setattr",
    "slice",
    "sorted",
    "staticmethod",
    "str",
    "sum",
    "super",
    "tuple",
    "type",
    "vars",
    "zip",
    "__import__",
    "ZeroDivisionError",
    "ValueError",
    "AssertionError",
    "AttributeError",
    "EOFError",
    "FloatingPointError",
    "GeneratorExit",
    "ImportError",
    "IndexError",
    "KeyError",
    "KeyboardInterrupt",
    "MemoryError",
    "NameError",
    "NotImplementedError",
    "OSError",
    "OverflowError",
    "ReferenceError",
    "RuntimeError",
    "StopIteration",
    "SyntaxError",
    "IndentationError",
    "TabError",
    "SystemError",
    "SystemExit",
    "TypeError",
    "UnboundLocalError",
    "UnicodeError",
    "UnicodeEncodeError",
    "UnicodeDecodeError",
    "UnicodeTranslateError",
    "and",
    "or",
    "if",
    "in",
    "elif",
    "else",
    "with",
    "open",
    "append",
    "update",
    "readlines",
    "startswith",
    "endswith",
    "keys",
    "values",
    "pop",
    "join",
    "strip",
]


# Keep this in alphabetical other , you b*st*rd

KEYWORDS_FTN = {
    "abs": "other",
    "aimag": "other",
    "aint": "other",
    "all": "other",
    "allocatable": "other",
    "allocate": "other",
    "allocated": "other",
    "anint": "other",
    "any": "other",
    "asin": "other",
    "assign": "other",
    "assignment": "other",
    "associate": "nesting",
    "atan": "other",
    "atan2": "other",
    "backspace": "other",
    "bind": "other",
    "bit_size": "other",
    "block": "nesting",
    "blockdata": "structure",
    "btest": "other",
    "call": "other",
    "case": "other",
    "ceiling": "other",
    "char": "other",
    "character": "other",
    "close": "other",
    "class": "other",
    "cmplx": "other",
    "common": "other",
    "conjg": "other",
    "contains": "other",
    "continue": "other",
    "cos": "other",
    "cosh": "other",
    "count": "other",
    "cshift": "other",
    "cycle": "other",
    "data": "other",
    "date_and_time": "other",
    "deallocate": "other",
    "digits": "other",
    "dim": "other",
    "dimension": "other",
    "do": "nesting",
    "dot_product": "other",
    "dprod": "other",
    "dshiftl": "other",
    "dshiftr": "other",
    "else": "other",
    "elseif": "other",
    "elsewhere": "other",
    "end": "other",
    "endfile": "other",
    "entry": "other",
    "enum": "nesting",
    "eor": "other",
    "epsilon": "other",
    "equivalence": "other",
    "erf": "other",
    "erfc": "other",
    "etime": "other",
    "exit": "other",
    "exp": "other",
    "exponent": "other",
    "extends": "other",
    "extends_type_of": "other",
    "external": "other",
    "floor": "other",
    "forall": "nesting",
    "format": "other",
    "fraction": "other",
    "function": "structure",
    "function_elemental": "structure",  # Created to simplify struct analysis
    "function_pure": "structure",  # Created to simplify struct analysis
    "function_recursive": "structure",  # Created to simplify struct analysis
    "gamma": "other",
    "go": "other",
    "goto": "other",
    "huge": "other",
    "iachar": "other",
    "iall": "other",
    "iand": "other",
    "iany": "other",
    "ibclr": "other",
    "ibits": "other",
    "ibset": "other",
    "ichar": "other",
    "idate": "other",
    "idim": "other",
    "idint": "other",
    "idnint": "other",
    "ieor": "other",
    "ierrno": "other",
    "if": "nesting",
    "ifix": "other",
    "implicit": "other",
    "index": "other",
    "inquire": "other",
    "int": "other",
    "integer": "other",
    "intent": "other",
    "interface": "structure",
    "interface_abstract": "structure",  # Created to simplify struct analysis
    "intrinsic": "other",
    "ior": "other",
    "iparity": "other",
    "irand": "other",
    "is_iostat_end": "other",
    "is_iostat_eor": "other",
    "ishft": "other",
    "ishftc": "other",
    "kind": "other",
    "lbound": "other",
    "lcm": "other",
    "len": "other",
    "len_trim": "other",
    "lge": "other",
    "lgt": "other",
    "lle": "other",
    "llt": "other",
    "log": "other",
    "log10": "other",
    "log_gamma": "other",
    "logical": "other",
    "maskl": "other",
    "maskr": "other",
    "matmul": "other",
    "max": "other",
    "maxexponent": "other",
    "maxloc": "other",
    "maxval": "other",
    "merge": "other",
    "merge_bits": "other",
    "min": "other",
    "minexponent": "other",
    "minloc": "other",
    "minval": "other",
    "mod": "other",
    "module": "structure",
    "modulo": "other",
    "namelist": "other",
    "nearest": "other",
    "norm2": "other",
    "not": "other",
    "null": "other",
    "nullify": "other",
    "nworkers": "other",
    "only": "other",
    "open": "other",
    "operator": "other",
    "optional": "other",
    "pack": "other",
    "parameter": "other",
    "parity": "other",
    "pause": "other",
    "pointer": "other",
    "popcnt": "other",
    "poppar": "other",
    "precision": "other",
    "present": "other",
    "print": "other",
    "private": "other",
    "procedure": "other",
    "product": "other",
    "program": "structure",
    "public": "other",
    "radix": "other",
    "range": "other",
    "read": "other",
    "real": "other",
    "recursive": "other",
    "repeat": "other",
    "reshape": "other",
    "result": "other",
    "return": "other",
    "rewind": "other",
    "rrspacing": "other",
    "save": "other",
    "scale": "other",
    "scan": "other",
    "select": "other",
    "select_case": "nesting",  # Created to simplify struct analysis
    "select_type": "nesting",  # Created to simplify struct analysis
    "selected_int_kind": "other",
    "selected_real_kind": "other",
    "sequence": "other",
    "set_exponent": "other",
    "shape": "other",
    "shifta": "other",
    "shiftl": "other",
    "shiftr": "other",
    "sign": "other",
    "sin": "other",
    "sinh": "other",
    "size": "other",
    "spacing": "other",
    "spread": "other",
    "sqrt": "other",
    "stop": "other",
    "storage_size": "other",
    "subroutine": "structure",
    "subroutine_elemental": "structure",  # Created to simplify struct analysis
    "subroutine_recursive": "structure",  # Created to simplify struct analysis
    "sum": "other",
    "system_clock": "other",
    "tan": "other",
    "tanh": "other",
    "target": "other",
    "then": "other",
    "this_image": "other",
    "tiny": "other",
    "to": "other",
    "trailz": "other",
    "transfer": "other",
    "transpose": "other",
    "trim": "other",
    "type": "structure",
    "type_private": "structure",  # Created to simplify struct analysis
    "type_private_abstract": "structure",  # Created to simplify struct analysis
    "type_public": "structure",  # Created to simplify struct analysis
    "type_public_abstract": "structure",  # Created to simplify struct analysis
    "ubound": "other",
    "unpack": "other",
    "use": "other",
    "verify": "other",
    "where": "nesting",
    "while": "other",
    "write": "other",
    "xor": "other",
}

STRUCTURES_FTN = [key for key, value in KEYWORDS_FTN.items() if value == "structure"]
NESTS_FTN = [key for key, value in KEYWORDS_FTN.items() if value == "nesting"]
OTHERS_FTN = [key for key, value in KEYWORDS_FTN.items() if value == "other"]



from loguru import logger
from typing import List, Tuple
from math import log2,sqrt

OPERATORS_COMMON = [
    '"',",",
    "(","[","{",
    ",",".",":",";",
    "+","/","-","*","=",
    "==","!=",">",">=","<","<=",
    "*=","+=","-=","/=",
    "&"]

OPERATORS_HIDDEN = [
    ')',"]","}"
   ]
EPS=1e-12
# def show_freqs(list):
#     #my_list = ['apple', 'banana', 'apple', 'orange', 'banana', 'apple']

#     # Create an empty dictionary to store frequencies
#     freq = {}

#     # Count the occurrences of each item
#     for item in list:
#         if item in freq:
#             freq[item] += 1
#         else:
#             freq[item] = 1

#     # Print the frequencies
#     for item, count in freq.items():
#         print(f'{item}: {count}')


def halstead_numbers(code: List[List[str]], intrinsics: List[str])-> Tuple[int,int,int,int]:
    """THis takes a list of tokenized code lines"""
    operators = [] 
    operands = [] 
    for line in code:
        for token in line:
            if token in OPERATORS_HIDDEN:
                continue
            elif token in intrinsics+OPERATORS_COMMON:
                operators.append(token)
            else:
                operands.append(token)
    # show_freqs(operators)
    # print("====", len(operators))
    # show_freqs(operands)
    # print("====", len(operands))
    
    unique_operators = set(operators)
    
    unique_operands = set(operands)
    #logger.info(sorted(unique_operators))
    #logger.info(sorted(unique_operands))
    return len(operators),len(operands),len(unique_operators),len(unique_operands)

def halstead_properties(optr_tot:int,oprd_tot:int,optr_set:int,oprd_set:int)-> Tuple[int,int,int,int]:
    """ see https://en.wikipedia.org/wiki/Halstead_complexity_measures
    """
    
    # print(optr_tot,oprd_tot,optr_set,oprd_set)
    
    length = optr_tot+oprd_tot
    vocabulary= optr_set+oprd_set
    #cal_est_prog_len = oprd_set*log(oprd_set)+oprd_set*log(oprd_set)
    volume = length*log2(vocabulary+EPS)
    difficulty=optr_set/2 * oprd_tot/(oprd_set+EPS)
    effort=volume*difficulty
    # print("vocabulary:",length)
    # print("length:",length)
    # print("volume:",volume)
    # print("difficulty:",volume)
    # print("effort:",volume)
    
    return round(volume,2), round(difficulty,2), round(effort,2)


def compute_cyclomatic_complexity(control_points,active_ctl_pts:list=None,passive_ctl_pts:list=None):
    """
    Computes the cyclomatic complexity based on nested list control points.

    Args:
        control_points (list): A nested list representing control flow points.

    Returns:
        int: The cyclomatic complexity.
    """
    if active_ctl_pts is None:
        active_ctl_pts=["if", "elif", "case"]
    if passive_ctl_pts is None:
        passive_ctl_pts=["else"]

    def count_complexity(points):
        complexity = 0
        for point in points:
            if isinstance(point, list):
                complexity += count_complexity(point)
            elif point in active_ctl_pts:
                complexity += 1
            elif point in passive_ctl_pts:
                pass
            else:
                logger.debug("")
                # the else is not an additional path
        return complexity

    # Start with a base complexity of 1 (one path through the code)
    base_complexity = 1
    total_complexity = base_complexity + count_complexity(control_points)
    return total_complexity


def compute_possible_paths(control_points:list):
    """
    Computes the possible_paths based on nested list control points.

    Args:
        control_points (list): A nested list representing control flow points.

    Returns:
        int: The possible paths
    """
    paths = 1

    for point in control_points:
        if isinstance(point, list):
            # Recursively count paths in nested lists
            nested_paths = compute_possible_paths(point)
            paths += nested_paths - 1
        elif point in ["if"]:
            paths = 2 * paths
        elif point in ["case", "elif"]:
            paths += 1
        else:
            pass
    return paths

def maintainability_index(halstead_volume:float, cyclomatic_complexity:float,loc:int, pct_comment:float=0)-> float:
    """ compute the maintainability index
    https://sourcery.ai/blog/maintainability-index/
    https://www.ecs.csun.edu/~rlingard/comp589/ColemanPaper.pdf
    """
    mi = 171 - 5.2 * log2(halstead_volume+EPS) - 0.23 * cyclomatic_complexity - 16.2 * log2(loc) + 50 * sqrt(2.46 * pct_comment)
    return round(mi,2)



"""
Module that gather the most common functions of unformat
"""

from __future__ import annotations
from typing import Tuple, List
from dataclasses import dataclass
from loguru import logger
import json


from tucan.string_utils import get_indent

CHAR_CMT_START = "\xaa"
CHAR_CMT_STOP = "\xbb"
CHAR_EOL = "\xcc"
CHAR_CMT_EOL = "\xdd"


@dataclass
class Statements:
    stmt: List[str] = None
    lines: List[Tuple[int, int]] = None

    def __str__(self):
        out = ""
        for line, (i1, i2) in zip(self.stmt, self.lines):
            out += f"({i1}-{i2})  {line}\n"
        out += "======================="
        out += f"Statements found : {len(self.stmt)}"
        return out

    def to_code(self):
        out = [f"{stmt}" for stmt in self.stmt]
        return out

    def to_nob(self):
        return [[st, lsp] for st, lsp in zip(self.stmt, self.lines)]

    def dump_json(self, newfile):
        logger.info(f"Data dumped to {newfile}")
        with open(newfile, "w") as fout:
            json.dump(self.to_nob(), fout, indent=2, sort_keys=True)

    def dump_code(self, newfile):
        with open(newfile, "w") as fout:
            fout.write("\n".join(self.to_code()))
        print(f"\n\nCleaned version dumped in {newfile}")


def new_stmts(code: List[str]) -> Statements:
    return Statements(code, [[i + 1, i + 1] for i, _ in enumerate(code)])


def remove_strings(code: List[str], string_char: str) -> List[str]:
    new_stmt = []

    for line in code:
        inside_string_context = False
        buffer = ""
        tmp_buffer = ""
        for char in line:
            if char == string_char:
                if not inside_string_context:
                    inside_string_context = True
                else:
                    buffer += f"{char}___"
                    inside_string_context = False
                    tmp_buffer = ""
                # buffer+=char

            if not inside_string_context:
                buffer += char
            else:
                tmp_buffer += char

        if (
            inside_string_context
        ):  # if the string context is not ended ath the end of the line, put the skipped chars
            buffer += tmp_buffer

        new_stmt.append(buffer)
    return new_stmt


def rm_parenthesis_content(
    buffer: str, lchar: str = "(", rchar: str = ")", greedy=True
) -> str:
    """Remove the content of parenthesis"""
    lvl = 0
    out = ""
    for char in buffer:
        if lvl == 0:
            out += char
        if char == lchar:
            lvl += 1

        if char == rchar:
            lvl -= 1

    if greedy:
        out = out.replace(lchar, "")
        out = out.replace(rchar, "")
    return out


def clean_blanks(stmts: Statements) -> Statements:
    """
    Perform the first pass to strip comments and blank lines.

    Args:
        code (List[str]): List of code lines.

    Returns:
        List[Tuple[str, int]]: List of code lines without comments and blank lines.
    """
    new_stmt = []
    new_lines = []
    for line, lines_span in zip(stmts.stmt, stmts.lines):
        if line.strip() == "":  # Blank lines
            continue

        new_stmt.append(line)
        new_lines.append(lines_span)

    # Add final newline if missing. Useful to finish cleanly the parsing
    if new_stmt:
        if new_stmt[-1] != "":
            new_stmt.append("")
            last_line = new_lines[-1][0]
            new_lines.append([last_line + 1, last_line + 1])

    return Statements(new_stmt, new_lines)


def clean_inline_comments(stmts: Statements, symbol: str = "#") -> Statements:
    """ """
    new_stmt = []
    new_lines = []
    for line, lines_span in zip(stmts.stmt, stmts.lines):
        if symbol in line:
            line = line.split(symbol)[0].rstrip()
            if line.strip() == "":
                continue
        new_stmt.append(line)
        new_lines.append(lines_span)

    return Statements(new_stmt, new_lines)


def clean_pure_comments(stmts: Statements, symbol: str = "#") -> Statements:
    """ """
    new_stmt = []
    new_lines = []
    for line, lines_span in zip(stmts.stmt, stmts.lines):
        if line.startswith(symbol):
            continue
        new_stmt.append(line)
        new_lines.append(lines_span)
    return Statements(new_stmt, new_lines)


def split_multi_statement_lines(stmts: Statements) -> Statements:
    """Split statements on ;"""
    new_stmt = []
    new_lines = []
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if ";" not in line:
            new_stmt.append(line)
            new_lines.append([lstart, lend])
        else:
            indent = get_indent(line)
            for stmt in line.split(";"):
                new_stmt.append(indent + stmt.strip())
                new_lines.append([lstart, lend])
    return Statements(new_stmt, new_lines)


def align_multiline_blocks(
    stmts: Statements, markup_in: str, markup_out: str
) -> Statements:
    merge_ = []
    i_in = None
    for i, line in enumerate(stmts.stmt):
        if markup_in in line:
            if i_in is not None:
                logger.info(f"Multiple start of continuation block at line {i}")
            i_in = i  # mark this line in all cases
        if markup_out in line:
            if i_in == i:
                pass  # no merge required actually
                i_in = None
            elif i_in is not None:
                merge_.append((i_in, i))  # store a merging range
                i_in = None
            else:
                logger.info(f"Spurious end of continuation block at line {i}")
    # logger.critical(merge_)

    new_stmt = stmts.stmt
    new_lines = stmts.lines

    def replace_slice_(lst, start, end, new_item):
        return lst[:start] + [new_item] + lst[end + 1 :]

    for i_in_, i_out_ in merge_:
        tmp_line = [stmts.lines[i_in_][0], stmts.lines[i_out_][1]]

        tmp_stmt = get_indent(stmts.stmt[i_in_]) + " ".join(
            [item.lstrip() for item in stmts.stmt[i_in_ : i_out_ + 1]]
        )

        new_stmt = replace_slice_(new_stmt, i_in_, i_out_, tmp_stmt)
        new_lines = replace_slice_(new_lines, i_in_, i_out_, tmp_line)

    return Statements(new_stmt, new_lines)


def strip_c_comment(line: str, fortran: bool = False) -> str:
    """
    Fortran=True :  equivalent to cpp --traditional
    """
    
    cline = []
    read = True
    _line = (
        line.replace("//", CHAR_CMT_EOL)
        .replace("/*", CHAR_CMT_START)
        .replace("*/", CHAR_CMT_STOP)
    )
    
    for char in _line:
        if char == CHAR_EOL:
            cline.append(char)
            if read is None:
                read = True
        elif char == CHAR_CMT_START and read is True:
            read = False  # Multiline comment start
        elif char == CHAR_CMT_STOP and read is False:
            read = True  # Multiline comment end
        elif char == CHAR_CMT_EOL and read:
            if fortran :
                cline.append(char)  # Keep the comment marker for Fortran, // is used for string concat
            else:
                read = None  # Single-line comment start
        elif read is True:
            cline.append(char)

    # Join the list into a string and replace markers
    return ''.join(cline).replace(CHAR_CMT_EOL, "//")

    # cline = ""
    # read = True
    # _line = (
    #     line.replace("//", CHAR_CMT_EOL)
    #     .replace("/*", CHAR_CMT_START)
    #     .replace("*/", CHAR_CMT_STOP)
    # )
    # for char in _line:
    #     if char == CHAR_EOL:
    #         cline += char
    #         if read == None:
    #             read = True
    #     elif char == CHAR_CMT_START and read is not None:
    #         read = False  # Read False if multiline Comment
    #     elif char == CHAR_CMT_STOP and read is not None:
    #         read = True
    #     elif char == CHAR_CMT_EOL and read is True:
    #         if not fortran:
    #             read = None  # Read None if single line comment
    #         else:
    #             cline += char  # Skip this in Fortran
    #     else:
    #         if read is True:
    #             cline += char

    # cline = cline.replace(CHAR_CMT_EOL, "//")
    # return cline


def clean_c_comments(lines: List[str], fortran: bool = False) -> List[str]:
    if not lines:
        return []
    raw = CHAR_EOL.join(lines)
    out = strip_c_comment(raw, fortran=fortran).split(CHAR_EOL)
    assert len(lines) == len(out)
    return out


def read_lines_with_encoding(filename: str) -> list:
    """Sacres francais"""
    try:
        with open(filename, "r", encoding="utf8") as fin:
            lines = fin.read().split("\n")
    except UnicodeDecodeError as e:
        with open(filename, "r", encoding="latin1") as fin:
            lines = fin.read().split("\n")

    return lines


def align_end_continuations(stmts: Statements, cont_char: str) -> Statements:
    """Align continuation with amprsand

    For example

     FORMAT("Lorem ipsum &
     sic hamet")

    """
    new_stmt = []
    new_lines = []
    last_line = ""
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if last_line.strip().endswith(cont_char):
            new_stmt[-1] = last_line[:-1] + " " + line.strip().lstrip(cont_char)
            new_lines[-1][1] = lend
        else:
            new_stmt.append(line)
            new_lines.append([lstart, lend])

        last_line = new_stmt[-1]
    return Statements(new_stmt, new_lines)



"""
Module that gather the most common functions of imports
"""


def imports_summary_str(imports_dict: dict) -> str:
    """
    Create a string to output the imports result for a single file.

    Args:
        imports_dict (dict): Imports found in a file.

    Returns:
        str: String with file name and its various imports.
    """
    out = []
    for file in imports_dict.keys():
        out.append(f"{file} :")
        if imports_dict[file]:
            for imported_from in imports_dict[file].keys():
                list_str = "\n       - " + "\n       - ".join(
                    imports_dict[file][imported_from]
                )
                out.append(f"    Functions imported from {imported_from} :{list_str}")

    return "\n".join(out)



import re, json
from typing import Tuple, List
from loguru import logger
from tucan.unformat_common import Statements
from tucan.struct_common import (
    new_buffer_item,
    new_stack_item,
    struct_from_stack,
    struct_augment,
    list2pathref,
)
from tucan.string_utils import find_words_before_left_parenthesis_noregexp
from tucan.kw_lang import KEYWORDS_C
from tucan.string_utils import tokenize
from tucan.complexities import (
    halstead_numbers,
    halstead_properties,
    compute_cyclomatic_complexity,
    compute_possible_paths,
)

from  tucan.tucanexceptions import TucanCtrlPtsError


def extract_struct_c(stmts: Statements, verbose: bool) -> dict:
    """Main calls to build structure form statements

    statements is the output of tucan.unformat_c.unformat_c
    """

    clean_code = stmts.to_code()
    all_structs = _extract_on_cleaned_c(stmts, verbose=verbose)
    all_structs = struct_augment(
        all_structs,
        clean_code,
        find_callables_c,
        compute_complexities_c,
        compute_cst_ccpp,
    )
    return all_structs


def resolve_scope_resolution(path, name) -> str:
    """find parent if any specified in the name"""
    if "." not in name:
        return None
    parent_path = path + name.split(".")
    in_parent = ".".join(parent_path[:-1])
    # logger.success(f"{name}|->|{in_parent}")
    return in_parent


def _extract_on_cleaned_c(stmts: Statements, verbose: bool = False) -> dict:
    """Extract structure from cleaned statements."""
    buffer = []
    stack = []
    path = []

    stat_idx = 0

    level = 0
    stack_level = [None]
    contains_dict = {}

    for line, (line_idx1, line_idx2) in zip(stmts.stmt, stmts.lines):
        stat_idx += 1
        part = ""

        if "###===" in line:
            type_, name, comment = _parse_type_name_c(line, line_idx1)
            name = name.replace("::", ".")
            # in_parent = resolve_scope_resolution(path,name)
            path.append(name)
            # if in_parent is not None:
            #     if in_parent not in contains_dict:
            #         contains_dict[in_parent] = [list2pathref(path)]
            #     else:
            #         contains_dict[in_parent].append(list2pathref(path))

            stack_level.append(level)
            buffer.append(
                new_buffer_item(
                    type_=type_,
                    path=path,
                    name=name,
                    first_line=line,
                    line_idx=line_idx1,
                    statement_idx=stat_idx,
                    comment=comment,
                    verbose=verbose,
                )
            )
            part = ""

        for char in line:
            part += char
            if char == "{":
                level += 1
            if char == "}":
                level -= 1
                if level == stack_level[-1]:
                    last_buff = buffer[-1]
                    stack.append(
                        new_stack_item(
                            last_buff, line_idx2, stat_idx, line, verbose=verbose
                        )
                    )
                    path.pop(-1)
                    buffer.pop(-1)
                    stack_level.pop(-1)

        # add contains links

    # logger.success(f"Contains {json.dumps(contains_dict, indent=4)}")
    # for stack_item in stack:
    # #    logger.success(f"Trying {stack_item.name}")
    #     if list2pathref(stack_item.path) in contains_dict:
    # #        logger.success(f"Adding contains to {stack_item.name}")
    #         stack_item.contains = contains_dict[list2pathref(stack_item.path) ]

    return struct_from_stack(
        stack,
        main_types=[
            #        "program ",
            #        "module ",
            #        "interface ",
            "int",
            "double",
            "char",
            "float",  # function
            "void",  # subroutine
            "struct",
            "enum",
            "class",
            "template",
            "namespace",  # "userdef_type",
            "pointer",
        ],
    )


def _parse_type_name_c(line: str, line_idx: int) -> Tuple[str, str, str]:
    """expect a lowercase stripped line
    takes the second word as the name

    :: scope resolution is replaced by . to avoid unwanted tokenization of names
    """

    if line.strip() == "":
        return None, None
    tokens = tokenize(line)

    if tokens[0] in ["for", "if", "else", "switch"]:
        return tokens[0], tokens[0] + str(line_idx + 1), None

    if tokens[0] in ["template"]:
        name_, type_, comment = read_template_name_and_type(tokens)
        return "template", name_, type_ + ";" + comment

    try:
        what, who = tokens[0], tokens[1]
    except IndexError:
        msgerr = f"Could not get name from line ({line_idx}){line}"
        raise RuntimeError(msgerr)
    return what, who, None


##### Main structs


def find_callables_c(tokenized_code: List[list]) -> list:
    """Find callables in c"""
    candidates = []
    for tokens in tokenized_code:
        candidates.extend(find_words_before_left_parenthesis_noregexp(tokens))
    matches = [
        cand.strip().replace("::", ".")
        for cand in set(candidates)
        if cand not in KEYWORDS_C
    ]
    # logger.critical(matches)
    return sorted(matches)  # Must be sorted for testing


def build_c_ctrl_points(code: List[list])-> list:
    """build the nested list of control points"""
    listpath=[]
    root_path=[]
    increment_control_pts = ["if",]
    constant_control_pts = ["case", "default"]

    # get exhaustive list of levels
    def _update_current_lvl(current_lvl,tokens):
        new_lvl = current_lvl
        for token in tokens:
            if token == "{":
                new_lvl +=1
            if token == "}":
                new_lvl -=1
        return new_lvl

    # build level information
    last_lvl=0
    levels = [last_lvl]    
    current_lvl = _update_current_lvl(0,code[0])
    for tokens_line in code[1:]:
        levels.append(current_lvl)
        current_lvl =  _update_current_lvl(current_lvl,tokens_line)
        
    # built control list (similiar to python)
    last_lvl=levels[0]
    last_token_line=code[0]
    for current_lvl,tokens_line in zip(levels[1:],code[1:]):
        #print("???",current_lvl,tokens_line)
    
        try: 
            anchor = last_token_line[0]
        except IndexError:
            anchor = None

        if current_lvl > last_lvl:
            root_path.append(anchor)
            listpath.append(root_path.copy())
            #print("+++",root_path)
        
        if tokens_line[0] in constant_control_pts:
            cst_ref= root_path.copy()+[ tokens_line[0]]
            listpath.append(cst_ref)
            #print("===",cst_ref)
        
        if current_lvl<last_lvl:
            while current_lvl<last_lvl:
                try:
                    root_path.pop(-1)
                except IndexError:
                    msg_err="Error while building controlpoints."
                    logger.warning(msg_err)
                    raise TucanCtrlPtsError(msg_err)
                last_lvl-=1
        
        last_lvl=current_lvl
        last_token_line=tokens_line

    # # remove non-control-points levels  
    # for path in  listpath:
    #     print("raw :",path)


    clean_list_path=[]   
    for i,path in enumerate(listpath):
        purged =  [item for item in path if item in increment_control_pts+constant_control_pts]
        if purged == [] :
            continue
        if not clean_list_path:
            clean_list_path.append(purged.copy())
            continue

        if purged[-1] in constant_control_pts:
            clean_list_path.append(purged.copy())
        elif purged != clean_list_path[-1]:
            clean_list_path.append(purged.copy())
           
    # for path in  clean_list_path:
    #     print("clean:",path)

    ctrl_points = [[]]
    for path in clean_list_path:
        if path == []:
            continue
        inserter=ctrl_points
        for _ in path:
            inserter=inserter[-1]
        inserter.extend([path[-1],[]])
    ctrl_points=ctrl_points[0]
 
    return ctrl_points


def compute_complexities_c(indents_code: List[int], tokenized_code: List[list]) -> int:
    """Count decision points (if, else if, do, select, etc.)"""
    avg_indent = sum(indents_code)/len(indents_code)/2
    ctrls_pts = build_c_ctrl_points(tokenized_code)
    mccabe = compute_cyclomatic_complexity(ctrls_pts)
    possible_paths = compute_possible_paths(ctrls_pts)
    volume, difficulty, effort =  halstead_properties(*halstead_numbers(tokenized_code, KEYWORDS_C))
    time_to_code = int(effort/18)

    return round(avg_indent,2),mccabe,possible_paths,volume, difficulty,time_to_code


# def compute_ccn_approx_c(code: list) -> int:
#     """Count decision points (if, else if, do, select, etc.)"""
#     decision_points = re.findall(
#         r"(?i)(if |else |for |case |default )", "\n".join(code)
#     )
#     complexity = len(decision_points) + 1
#     return complexity


def read_template_name_and_type(tokens) -> Tuple[str, str, str]:
    """Extract template name and typology form the declaration"""
    # read template defs

    comment = ""
    if "static" in tokens:
        comment += "static"

    tokens = [token for token in tokens if token not in ["static", "inline"]]

    assert tokens[0] == "template"
    idx = 1

    if tokens[idx] == "<":
        content0, idx = read_chevrons(tokens, start=idx)
        comment += "<" + ",".join(content0) + ">"
    else:
        content0 = []
    idx += 1

    # read template first element (type for a function, name for a constructor)
    tok_1 = tokens[idx]
    idx += 1
    if tokens[idx] == "(":  # assumes it is a constructor
        return tok_1, "constructor", comment
    else:
        type_ = tok_1

    if tokens[idx] == "<":
        content, idx = read_chevrons(tokens, start=idx)
        if content:
            type_ = "<" + " ".join(content) + ">" + type_

    # read template name
    name = tokens[idx]
    if name == "&":
        idx += 1
        name = "&" + tokens[idx]
    # if content0:
    #     name += "."+"|".join(content0)

    idx += 1
    if tokens[idx] == "<":
        content, idx = read_chevrons(tokens, start=idx)
        if content:
            name = name + ".<" + " ".join(content) + ">"
    else:
        content = ""

    return name, type_, comment


def read_chevrons(list_, start=0) -> Tuple[list, int]:
    """Read chevrons declarations in cpp code, especially in the context of templates"""
    assert list_[start] == "<"
    end_idx = list_.index(">", start)
    idx = start
    content = []
    while idx < end_idx:
        if list_[idx] == "=":
            idx += 2
            continue
        if list_[idx] not in ["<", "typename", ",", "=", "..."]:
            content.append(list_[idx])
        idx += 1

    joined_content = []
    join = False
    for item in content:
        if item in ["::"]:
            join = True
        else:
            if join:
                joined_content[-1] += "::" + item
                join = False
            else:
                joined_content.append(item)

    return joined_content, idx


def compute_cst_ccpp(type_: str) -> int:
    """State the structural complexity of a code

    in Short, the average nb. of time we re-read the element.
    It does NOT means it's bad practice
    It just means more read time for the reader to understand the code"""

    cst_ = {
        "int": 1,
        "double": 1,
        "char": 1,
        "float": 1,
        "void": 2,
        "namespace": 2,
        "pointer": 2,
        "enum": 2,
        "struct": 3,
        "class": 4,
        "template": 8,
    }
    if type_ not in cst_:
        logger.warning(f"Type {type_} not present in conversion list ")
        return 1

    return cst_.get(type_, 1)



import sys
from loguru import logger

def startlog(verbose: bool):
    """General logging function of PACKAGE

    """
    format_verbose = "<level>{level}</level> - <blue>{file}:{function}:{line}</blue> - <level>{message}</level>"
    format_normal = "<level>{message}</level>"
    logger.remove()    
    if verbose:
        logger.add(sys.stdout, format=format_verbose,level="TRACE",backtrace=True,diagnose=True)
    else:
        logger.add(sys.stdout, format=format_normal, level="INFO",backtrace= False)
        
    


"""Module that aims to analyze a whole package based 
on the other unitary function of the package"""

from loguru import logger
from pathlib import Path
from tucan.unformat_main import unformat_main
from tucan.struct_main import struct_main
from tucan.imports_main import imports_of_file
from tucan.travel_in_package import clean_extensions_in_paths
from tucan.tucanexceptions import TucanError

def run_full_analysis(files: dict) -> dict:
    """
    Gather the data associated to the functions and the imports within a file

    Args:
        files (dict): key: short_name , value: absolute paths

    Returns:
        dict: _description_
    """

    full_analysis = {}

    skipped_files = []
    unrecoverable_files = []
    for file, path_ in files.items():
        imps,_ = imports_of_file(path_)
        full_analysis[file] = {"imports": imps}
        analysis = struct_main(path_)
        if analysis == {}:
            skipped_files.append(file)
        if analysis == None:
            unrecoverable_files.append(file)
            analysis = {}

        full_analysis[file].update(analysis)
    logger.success("Analyze completed.")
    if unrecoverable_files:
        logger.warning("Some files could not be parsed correctly")
        for f_ in unrecoverable_files + skipped_files:
            print(" - ", f_)

    return full_analysis


def run_unformat(clean_paths: list) -> dict:
    """
    Gather the unformated version of code files within a dict.

    Args:
        clean_paths (list): List of cleaned paths.

    Returns:
        dict: File path as key, item as a list of lines with their line number span
    """
    statements = {}
    for file in clean_paths:
        statements[file] = unformat_main(file).to_nob()

        nbr_of_stmt = 0
        if statements[file]:
            nbr_of_stmt = len(statements[file])
        logger.info(f"Found {nbr_of_stmt} statements for {file}")

    return statements


def run_struct(clean_paths: list, ignore_errors:bool=True) -> dict:
    """
    Gather the data associated to the functions within a file.

    Args:
        clean_paths (list): List of cleaned paths.

    Returns:
        dict: File path as key, item as dict with function names and their data (NLOC, CCN, etc.)
    """
    full_struct = {}
    files = []

    for path_ in clean_paths:
        if not Path(path_).is_dir():
            files.append(path_)

    files = clean_extensions_in_paths(files)
    for file in files:
        try:
            full_struct[file] = struct_main(file)
        except TucanError:
            logger.warning(f"Struct analysis failed on {file}")
            if ignore_errors:
                full_struct[file] = {}
            else:
                userinput = input("Would you like to continue (y/n)?")
                if userinput == "y":
                    full_struct[file] = {}
                else:
                    raise # raise previous error

        total_stmts = 0
        for _, data in full_struct[file].items():
            total_stmts += data["ssize"]
        logger.info(f"Found {total_stmts} statements for {file}")

    return full_struct